"""Optical components representing physical beamsplitters."""

import logging
import types
import numpy as np

from finesse.exceptions import TotalReflectionError
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index

from finesse.components.general import InteractionType, DOFDefinition, NoiseType
from finesse.components.surface import Surface
from finesse.components.node import NodeDirection, NodeType, Node

LOGGER = logging.getLogger(__name__)


@float_parameter("R", "Reflectivity", validate="_check_R", setter="set_RTL")
@float_parameter("T", "Transmission", validate="_check_T", setter="set_RTL")
@float_parameter("L", "Loss", validate="_check_L", setter="set_RTL")
@float_parameter("phi", "Phase", units="degrees")
@float_parameter(
    "alpha",
    "Angle of incidence (-90 <= alpha <= 90)",
    units="degrees",
    is_geometric=True,
    changeable_during_simulation=False,
)
@float_parameter(
    "Rcx",
    "Radius of curvature (x)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter(
    "Rcy",
    "Radius of curvature (y)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter("xbeta", "Yaw misalignment", units="radians")
@float_parameter("ybeta", "Pitch misalignment", units="radians")
class Beamsplitter(Surface):
    """The beamsplitter component represents a thin dielectric surface with associated
    properties such as reflectivity, tuning, and radius of curvature. It has four
    optical ports `p1`, `p2`, `p3`, and `p4` which decsribe the four beams incident on
    either side of this surface. `p1` and `p2` are on side 1 and `p3` and `p4` are on
    side 2.

    It also has a mechanical port `mech` which has nodes for longitudinal, yaw, and pitch motions.
    These mechanical nodes are purely for exciting small signal oscillations of the mirror.
    Static offsets in longitudinal displacements are set by the `phi` parameter (in units of degrees),
    misalignments in yaw by the `xbeta` parameter, and pitch the `ybeta` parameter. Macroscopic angle
    of incidence of the beamsplitter is set by the ``alpha`` parameter.

    Beamsplitters physically operate the same as mirror components, except for the non-normal angle of incidence
    option.

    Parameters
    ----------
    name : str
        Name of newly created beamsplitter.

    R : float, optional
        Reflectivity of the beamsplitter.

    T : float, optional
        Transmissivity of the beamsplitter.

    L : float, optional
        Loss of the beamsplitter.

    phi : float, optional
        Microscopic tuning of the beamsplitter (in degrees).

    alpha : float, optional
        Angle of incidence (in degrees)

    Rc : float, optional
        Radius of curvature (in metres); defaults to ``numpy.inf``
        to indicate a planar surface.

    xbeta, ybeta : float, optional
        Angle of misalignment in the yaw plane (xbeta) and pitch
        (ybeta), respectively (in radians); defaults to `0`.
    """

    def __init__(
        self, name, R=None, T=None, L=None, phi=0, alpha=0, Rc=np.inf, xbeta=0, ybeta=0
    ):
        super().__init__(name, R, T, L, phi, Rc, xbeta, ybeta)
        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p3", NodeType.OPTICAL)
        self.p3._add_node("i", NodeDirection.INPUT)
        self.p3._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p4", NodeType.OPTICAL)
        self.p4._add_node("i", NodeDirection.INPUT)
        self.p4._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings => reflections
        self._register_node_coupling(
            "P1i_P2o", self.p1.i, self.p2.o, interaction_type=InteractionType.REFLECTION
        )
        self._register_node_coupling(
            "P2i_P1o", self.p2.i, self.p1.o, interaction_type=InteractionType.REFLECTION
        )
        self._register_node_coupling(
            "P3i_P4o", self.p3.i, self.p4.o, interaction_type=InteractionType.REFLECTION
        )
        self._register_node_coupling(
            "P4i_P3o", self.p4.i, self.p3.o, interaction_type=InteractionType.REFLECTION
        )

        # optic to optic couplings => transmissions
        self._register_node_coupling(
            "P1i_P3o",
            self.p1.i,
            self.p3.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P4o",
            self.p2.i,
            self.p4.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P3i_P1o",
            self.p3.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P4i_P2o",
            self.p4.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

        # mirror motion couplings
        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", NodeDirection.OUTPUT)
        self.mech._add_node("yaw", NodeDirection.OUTPUT)
        self.mech._add_node("pitch", NodeDirection.OUTPUT)
        self.mech._add_node("F_z", NodeDirection.INPUT)
        self.mech._add_node("F_yaw", NodeDirection.INPUT)
        self.mech._add_node("F_pitch", NodeDirection.INPUT)

        # optic to motion couplings
        self._register_node_coupling("P1i_Fz", self.p1.i, self.mech.F_z)
        self._register_node_coupling("P2i_Fz", self.p2.i, self.mech.F_z)
        self._register_node_coupling("P3i_Fz", self.p3.i, self.mech.F_z)
        self._register_node_coupling("P4i_Fz", self.p4.i, self.mech.F_z)
        self._register_node_coupling("P1o_Fz", self.p1.o, self.mech.F_z)
        self._register_node_coupling("P2o_Fz", self.p2.o, self.mech.F_z)
        self._register_node_coupling("P3o_Fz", self.p3.o, self.mech.F_z)
        self._register_node_coupling("P4o_Fz", self.p4.o, self.mech.F_z)
        self._register_node_coupling("P1i_Fyaw", self.p1.i, self.mech.F_yaw)
        self._register_node_coupling("P2i_Fyaw", self.p2.i, self.mech.F_yaw)
        self._register_node_coupling("P3i_Fyaw", self.p3.i, self.mech.F_yaw)
        self._register_node_coupling("P4i_Fyaw", self.p4.i, self.mech.F_yaw)
        self._register_node_coupling("P1o_Fyaw", self.p1.o, self.mech.F_yaw)
        self._register_node_coupling("P2o_Fyaw", self.p2.o, self.mech.F_yaw)
        self._register_node_coupling("P3o_Fyaw", self.p3.o, self.mech.F_yaw)
        self._register_node_coupling("P4o_Fyaw", self.p4.o, self.mech.F_yaw)
        self._register_node_coupling("P1i_Fpitch", self.p1.i, self.mech.F_pitch)
        self._register_node_coupling("P2i_Fpitch", self.p2.i, self.mech.F_pitch)
        self._register_node_coupling("P3i_Fpitch", self.p3.i, self.mech.F_pitch)
        self._register_node_coupling("P4i_Fpitch", self.p4.i, self.mech.F_pitch)
        self._register_node_coupling("P1o_Fpitch", self.p1.o, self.mech.F_pitch)
        self._register_node_coupling("P2o_Fpitch", self.p2.o, self.mech.F_pitch)
        self._register_node_coupling("P3o_Fpitch", self.p3.o, self.mech.F_pitch)
        self._register_node_coupling("P4o_Fpitch", self.p4.o, self.mech.F_pitch)
        # motion to optic couplings: phase coupling on reflection
        self._register_node_coupling("Z_P1o", self.mech.z, self.p1.o)
        self._register_node_coupling("Z_P2o", self.mech.z, self.p2.o)
        self._register_node_coupling("Z_P3o", self.mech.z, self.p3.o)
        self._register_node_coupling("Z_P4o", self.mech.z, self.p4.o)
        self._register_node_coupling("yaw_P1o", self.mech.yaw, self.p1.o)
        self._register_node_coupling("yaw_P2o", self.mech.yaw, self.p2.o)
        self._register_node_coupling("yaw_P3o", self.mech.yaw, self.p3.o)
        self._register_node_coupling("yaw_P4o", self.mech.yaw, self.p4.o)
        self._register_node_coupling("pitch_P1o", self.mech.pitch, self.p1.o)
        self._register_node_coupling("pitch_P2o", self.mech.pitch, self.p2.o)
        self._register_node_coupling("pitch_P3o", self.mech.pitch, self.p3.o)
        self._register_node_coupling("pitch_P4o", self.mech.pitch, self.p4.o)

        # NOTE (pjj) temporarily moved after add_port, as validator depends on ports
        #            being present. Same fix as below.
        self.alpha = alpha

        self.__changing_check = set(
            (self.R, self.T, self.L, self.phi, self.alpha, self.xbeta, self.ybeta)
        )

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        self.dofs.z = DOFDefinition(self.phi, self.mech.z, 1)
        self.dofs.yaw = DOFDefinition(self.xbeta, self.mech.yaw, 1)
        self.dofs.pitch = DOFDefinition(self.ybeta, self.mech.pitch, 1)

    def get_adjacent_port(self, p):
        """Get the port adjacent (on the same side of the surface) as `p`."""
        if isinstance(p, Node):
            p = p.port

        adj_dict = {
            self.p1: self.p2,
            self.p2: self.p1,
            self.p3: self.p4,
            self.p4: self.p3,
        }
        if p not in adj_dict:
            raise ValueError(f"Port {p} does not belong to Beamsplitter {self.name}")

        return adj_dict[p]

    def _resymbolise_ABCDs(self):
        # -> reflections
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "y")
        self.__symbolise_ABCD(self.p3.i, self.p4.o, "x")
        self.__symbolise_ABCD(self.p3.i, self.p4.o, "y")
        self.__symbolise_ABCD(self.p4.i, self.p3.o, "x")
        self.__symbolise_ABCD(self.p4.i, self.p3.o, "y")
        # -> transmissions
        self.__symbolise_ABCD(self.p1.i, self.p3.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p3.o, "y")
        self.__symbolise_ABCD(self.p3.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p3.i, self.p1.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p4.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p4.o, "y")
        self.__symbolise_ABCD(self.p4.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p4.i, self.p2.o, "y")

    def __symbolise_ABCD(self, from_node, to_node, direction):
        if direction == "x":
            Rc = self.Rcx.ref
        else:
            Rc = self.Rcy.ref

        alpha = self.alpha.ref

        nr1 = refractive_index(from_node, symbolic=True)
        nr2 = refractive_index(to_node, symbolic=True)

        alpha1 = np.radians(alpha)
        # we get alpha2 from Snell's law
        if from_node.port is self.p3 or from_node.port is self.p4:
            sin_alpha2 = (nr2 / nr1) * np.sin(alpha1)
        else:
            sin_alpha2 = (nr1 / nr2) * np.sin(alpha1)

        if abs(float(sin_alpha2)) > 1:
            msg = (
                f"Total reflection in beam splitter {self.name} for "
                f"{from_node.full_name} -> {to_node.full_name} in plane {direction}."
            )
            tr = TotalReflectionError(msg, from_node, to_node)
            self._abcd_matrices[(from_node, to_node, direction)] = None, tr
            return

        alpha2 = np.arcsin(sin_alpha2)

        A = 1.0
        D = 1.0
        # reflection
        if self.interaction_type(from_node, to_node) == InteractionType.REFLECTION:
            if from_node.port is self.p1 or from_node.port is self.p2:
                if direction == "x":
                    C = -2 * nr1 / (Rc * np.cos(alpha1))
                else:
                    C = -2 * nr1 * np.cos(alpha1) / Rc
            else:
                if direction == "x":
                    C = 2 * nr2 / (Rc * np.cos(alpha2))
                else:
                    C = 2 * nr2 * np.cos(alpha2) / Rc

        # transmission
        else:
            cos_alpha1 = np.cos(alpha1)
            cos_alpha2 = np.cos(alpha2)
            delta_n = (nr2 * cos_alpha2 - nr1 * cos_alpha1) / (cos_alpha1 * cos_alpha2)
            C = delta_n / Rc

            if direction == "x":
                A = cos_alpha2 / cos_alpha1
                D = cos_alpha1 / cos_alpha2
                if from_node.port is self.p3 or from_node.port is self.p4:
                    A = 1 / A
                    D = 1 / D

        M_sym = np.array([[A, 0.0], [C, D]])

        key = (from_node, to_node, direction)
        # For beamsplitters the symbol nr can change when connected up to spaces
        # in a model so here we update the ABCD matrix entry
        if key in self._abcd_matrices:
            self._abcd_matrices[key] = M_sym, np.array(M_sym, dtype=np.float64)
        # Otherwise just register the new ABCD matrix as usual
        else:
            self.register_abcd_matrix(M_sym, (from_node, to_node, direction))

    @property
    def abcd12x(self):
        """Numeric ABCD matrix from port 1 to port 2 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(1, 2, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "x")

    @property
    def abcd12y(self):
        """Numeric ABCD matrix from port 1 to port 2 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(1, 2, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "y")

    @property
    def abcd21x(self):
        """Numeric ABCD matrix from port 2 to port 1 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(2, 1, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 1, "x")

    @property
    def abcd21y(self):
        """Numeric ABCD matrix from port 2 to port 1 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(2, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 1, "y")

    @property
    def abcd34x(self):
        """Numeric ABCD matrix from port 3 to port 4 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(3, 4, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(3, 4, "x")

    @property
    def abcd34y(self):
        """Numeric ABCD matrix from port 3 to port 4 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(3, 4, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(3, 4, "y")

    @property
    def abcd43x(self):
        """Numeric ABCD matrix from port 4 to port 3 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(4, 3, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(4, 3, "x")

    @property
    def abcd43y(self):
        """Numeric ABCD matrix from port 4 to port 3 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(4, 3, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(4, 3, "y")

    @property
    def abcd13x(self):
        """Numeric ABCD matrix from port 1 to port 3 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(1, 3, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 3, "x")

    @property
    def abcd13y(self):
        """Numeric ABCD matrix from port 1 to port 3 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(1, 3, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 3, "y")

    @property
    def abcd31x(self):
        """Numeric ABCD matrix from port 3 to port 1 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(3, 1, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(3, 1, "x")

    @property
    def abcd31y(self):
        """Numeric ABCD matrix from port 3 to port 1 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(3, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(3, 1, "y")

    @property
    def abcd24x(self):
        """Numeric ABCD matrix from port 2 to port 4 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(2, 4, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 4, "x")

    @property
    def abcd24y(self):
        """Numeric ABCD matrix from port 2 to port 4 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(2, 4, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 4, "y")

    @property
    def abcd42x(self):
        """Numeric ABCD matrix from port 4 to port 2 in the tangential plane.

        Equivalent to ``beamsplitter.ABCD(4, 2, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(4, 2, "x")

    @property
    def abcd42y(self):
        """Numeric ABCD matrix from port 4 to port 2 in the sagittal plane.

        Equivalent to ``beamsplitter.ABCD(4, 2, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(4, 2, "y")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
    ):
        r"""Returns the ABCD matrix of the beam splitter for the specified coupling.

        The matrices for transmission and reflection are different for
        the sagittal and tangential planes (:math:`M_s` and :math:`M_t`),
        as shown below.

        .. rubric:: Transmission

        .. _fig_abcd_bs_transmission:
        .. figure:: ../images/abcd_bst.*
            :align: center

        For the tangential plane (`direction = 'x'`),

        .. math::
            M_t = \begin{pmatrix}
                        \frac{\cos{\alpha_2}}{\cos{\alpha_1}} & 0 \\
                        \frac{\Delta n}{R_c} & \frac{\cos{\alpha_1}}{\cos{\alpha_2}}
                    \end{pmatrix},

        and for the sagittal plane (`direction = 'y'`),

        .. math::
            M_s = \begin{pmatrix}
                        1 & 0 \\
                        \frac{\Delta n}{R_c} & 1
                    \end{pmatrix},

        where :math:`\alpha_1` is the angle of incidence of the beam splitter and
        :math:`\alpha_2` is given by Snell's law (:math:`n_1\sin{\alpha_1} =
        n_2\sin{\alpha_2}`). The quantity :math:`\Delta n` is given by,

        .. math::
            \Delta_n = \frac{n_2 \cos{\alpha_2} - n_1 \cos{\alpha_1}}{
                \cos{\alpha_1} \cos{\alpha_2}
                }.

        If the direction of propagation is reversed such that the radius of curvature
        of the beam splitter is in this direction, then the elements :math:`A` and
        :math:`D` of the tangential matrix (:math:`M_t`) are swapped.

        .. rubric:: Reflection

        .. _fig_abcd_bs_reflection:
        .. figure:: ../images/abcd_bsr.*
            :align: center

        The reflection at the front surface of the beam splitter is given by,

        .. math::
            M_t = \begin{pmatrix}
                        1 & 0 \\
                        -\frac{2n_1}{R_c \cos{\alpha_1}} & 1
                    \end{pmatrix},

        for the tangential plane, and,

        .. math::
            M_s = \begin{pmatrix}
                        1 & 0 \\
                        -\frac{2n_1 \cos{\alpha_2}}{R_c} & 1
                    \end{pmatrix},

        for the sagittal plane.

        At the back surface :math:`R_c \rightarrow - R_c` and
        :math:`\alpha_1 \rightarrow - \alpha_2`.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.

        Raises
        ------
        tre : :class:`.TotalReflectionError`
            If total reflection occurs for the specified coupling - i.e. if :math:`\sin{\alpha_2} > 1.0`.
        """
        return super().ABCD(from_node, to_node, direction, symbolic, copy, retboth)

    def _get_workspace(self, sim):
        from finesse.components.modal.beamsplitter import (
            beamsplitter_carrier_fill,
            beamsplitter_signal_fill,
            beamsplitter_fill_qnoise,
            BeamsplitterWorkspace,
        )

        _, is_changing = self._eval_parameters()

        misaligned = self.xbeta.value != 0 or self.ybeta.value != 0
        refill = (
            (sim.signal and sim.model.fsig.f.is_changing)
            or sim.is_component_in_mismatch_couplings(self)
            or (misaligned and sim.trace_forest.contains_comp(self))
            or sim.carrier.any_frequencies_changing
            or (len(is_changing) and is_changing.issubset(self.__changing_check))
        )

        ws = BeamsplitterWorkspace(self, sim, refill)
        ws.carrier.add_fill_function(beamsplitter_carrier_fill, refill)
        ws.signal.add_fill_function(beamsplitter_signal_fill, refill)

        # Initialise the ABCD matrix memory-views
        if sim.is_modal:
            try:
                ws.abcd_p1p2_x = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)
                ws.abcd_p1p2_y = self.ABCD(self.p1.i, self.p2.o, "y", copy=False)
            except TotalReflectionError:
                raise

            try:
                ws.abcd_p2p1_x = self.ABCD(self.p2.i, self.p1.o, "x", copy=False)
                ws.abcd_p2p1_y = self.ABCD(self.p2.i, self.p1.o, "y", copy=False)
            except TotalReflectionError:
                raise

            try:
                ws.abcd_p3p4_x = self.ABCD(self.p3.i, self.p4.o, "x", copy=False)
                ws.abcd_p3p4_y = self.ABCD(self.p3.i, self.p4.o, "y", copy=False)
            except TotalReflectionError:
                raise

            try:
                ws.abcd_p4p3_x = self.ABCD(self.p4.i, self.p3.o, "x", copy=False)
                ws.abcd_p4p3_y = self.ABCD(self.p4.i, self.p3.o, "y", copy=False)
            except TotalReflectionError:
                raise

            ws.abcd_p1p3_x = self.ABCD(self.p1.i, self.p3.o, "x", copy=False)
            ws.abcd_p1p3_y = self.ABCD(self.p1.i, self.p3.o, "y", copy=False)
            ws.abcd_p3p1_x = self.ABCD(self.p3.i, self.p1.o, "x", copy=False)
            ws.abcd_p3p1_y = self.ABCD(self.p3.i, self.p1.o, "y", copy=False)
            ws.abcd_p2p4_x = self.ABCD(self.p2.i, self.p4.o, "x", copy=False)
            ws.abcd_p2p4_y = self.ABCD(self.p2.i, self.p4.o, "y", copy=False)
            ws.abcd_p4p2_x = self.ABCD(self.p4.i, self.p2.o, "x", copy=False)
            ws.abcd_p4p2_y = self.ABCD(self.p4.i, self.p2.o, "y", copy=False)

            alpha2 = np.arcsin((ws.nr1 / ws.nr2) * np.sin(np.deg2rad(self.alpha.value)))
            cos_alpha = np.cos(np.deg2rad(self.alpha.value))
            cos_alpha2 = np.cos(alpha2)

            ws.set_knm_info(
                "P1i_P2o",
                alpha=self.alpha,
                is_transmission=False,
                nr_from=ws.nr1,
                nr_to=ws.nr1,
                beta_x=self.xbeta,
                beta_x_factor=2,
                beta_y=self.ybeta,
                beta_y_factor=-2 * cos_alpha,
                abcd_x=ws.abcd_p1p2_x,
                abcd_y=ws.abcd_p1p2_y,
            )
            ws.set_knm_info(
                "P2i_P1o",
                alpha=self.alpha,
                is_transmission=False,
                nr_from=ws.nr1,
                nr_to=ws.nr1,
                beta_x=self.xbeta,
                beta_x_factor=2,
                beta_y=self.ybeta,
                beta_y_factor=-2 * cos_alpha,
                abcd_x=ws.abcd_p2p1_x,
                abcd_y=ws.abcd_p2p1_y,
            )
            ws.set_knm_info(
                "P3i_P4o",
                alpha=self.alpha,
                is_transmission=False,
                nr_from=ws.nr2,
                nr_to=ws.nr2,
                beta_x=self.xbeta,
                beta_x_factor=2,
                beta_y=self.ybeta,
                beta_y_factor=2 * cos_alpha2,
                abcd_x=ws.abcd_p3p4_x,
                abcd_y=ws.abcd_p3p4_y,
            )
            ws.set_knm_info(
                "P4i_P3o",
                alpha=self.alpha,
                is_transmission=False,
                nr_from=ws.nr2,
                nr_to=ws.nr2,
                beta_x=self.xbeta,
                beta_x_factor=2,
                beta_y=self.ybeta,
                beta_y_factor=2 * cos_alpha2,
                abcd_x=ws.abcd_p4p3_x,
                abcd_y=ws.abcd_p4p3_y,
            )
            ws.set_knm_info(
                "P1i_P3o",
                alpha=self.alpha,
                is_transmission=True,
                nr_from=ws.nr1,
                nr_to=ws.nr2,
                beta_x=self.xbeta,
                beta_x_factor=-(1 - ws.nr1 / ws.nr2),
                beta_y=self.ybeta,
                beta_y_factor=(1 - ws.nr1 / ws.nr2),
                abcd_x=ws.abcd_p1p3_x,
                abcd_y=ws.abcd_p1p3_y,
            )
            ws.set_knm_info(
                "P3i_P1o",
                alpha=self.alpha,
                is_transmission=True,
                nr_from=ws.nr2,
                nr_to=ws.nr1,
                beta_x=self.xbeta,
                beta_x_factor=-(1 - ws.nr2 / ws.nr1),
                beta_y=self.ybeta,
                beta_y_factor=-(1 - ws.nr2 / ws.nr1),
                abcd_x=ws.abcd_p3p1_x,
                abcd_y=ws.abcd_p3p1_y,
            )
            ws.set_knm_info(
                "P2i_P4o",
                alpha=self.alpha,
                is_transmission=True,
                nr_from=ws.nr1,
                nr_to=ws.nr2,
                beta_x=self.xbeta,
                beta_x_factor=-(1 - ws.nr1 / ws.nr2),
                beta_y=self.ybeta,
                beta_y_factor=(1 - ws.nr1 / ws.nr2),
                abcd_x=ws.abcd_p2p4_x,
                abcd_y=ws.abcd_p2p4_y,
            )
            ws.set_knm_info(
                "P4i_P2o",
                alpha=self.alpha,
                is_transmission=True,
                nr_from=ws.nr2,
                nr_to=ws.nr1,
                beta_x=self.xbeta,
                beta_x_factor=-(1 - ws.nr2 / ws.nr1),
                beta_y=self.ybeta,
                beta_y_factor=-(1 - ws.nr2 / ws.nr1),
                abcd_x=ws.abcd_p4p2_x,
                abcd_y=ws.abcd_p4p2_y,
            )

        if sim.signal:
            ws.signal.set_fill_noise_function(
                NoiseType.QUANTUM, beamsplitter_fill_qnoise
            )
        return ws



import re
import pytest
from finesse.script.exceptions import KatParsingError
from ....util import dedent_multiline


@pytest.fixture
def compiler(compiler):
    analysis = lambda a: None
    compiler.spec.register_analysis(
        "fake_analysis", {"setter": analysis, "getter": analysis}
    )
    return compiler


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                fake_analysis()
                fake_analysis()
                """
            ),
            "\nlines 1-2: duplicate analysis trees (combine with 'series' or 'parallel')\n"
            "-->1: fake_analysis()\n"
            "      ^^^^^^^^^^^^^\n"
            "-->2: fake_analysis()\n"
            "      ^^^^^^^^^^^^^",
            id="xaxis",
        ),
    ),
)
def test_multiple_analyses_invalid(compiler, script, error):
    with pytest.raises(KatParsingError, match=re.escape(error)):
        compiler.compile(script)



"""Gaussian beam related tools and utilities."""

import logging

import math
import numpy as np

from finesse.config import config_instance
from finesse.cymath.gaussbeam import transform_q
from finesse.cymath.complex import ceq
from finesse.cymath.homs import HGModeWorkspace
from finesse.symbols import Symbol, FUNCTIONS
from .utilities.misc import is_iterable


LOGGER = logging.getLogger(__name__)


def transform_beam_param(ABCD, q1, nr1=1, nr2=1):
    r"""Transforms a beam parameter `q1` by the matrix `ABCD`.

    The value of the new beam parameter returned is given by the
    Kogelnik transformation,

    .. math::
        q_2 = n_{r,2} \frac{A \frac{q_1}{n_{r,1}} + B}{ C \frac{q_1}{n_{r_1}} + D},

    where :math:`A`, :math:`B`, :math:`C`, :math:`D` are the
    corresponding elements of the `ABCD` matrix, and :math:`n_{r,1}`,
    :math:`n_{r,2}` are the refractive indices of the source and target
    media respectively.

    .. note::

        The method :meth:`.Model.propagate_beam` should be preferred over this in most cases,
        as it will automatically take into account refractive indices of spaces. With that
        method, you can simply specify the input and output nodes associated with the component
        you want to transform over and get back a :class:`.PropagationSolution` instance on
        which you can find the resulting beam parameter with :meth:`.PropagationSolution.q`.

        This function is more intended for array-calculations (i.e. where `q1` is an array
        of beam parameters or complex values) and cases where `ABCD` is a manually defined
        matrix.

    Parameters
    ----------
    ABCD : :class:`numpy.ndarray`
        A 2D numpy array of shape ``(2, 2)`` containing the ABCD matrix elements. Note
        that this can be a matrix of symbolic elements or a matrix of numbers.

    q1 : complex, :class:`.BeamParam`, symbolic, array-like
        A complex number, :class:`.BeamParam` instance or complex symbolic expression,
        representing the beam parameter to be transformed by `ABCD`. This can also be
        an array of beam parameters (which in turn can be symbolic or numeric again).

    nr1 : float, symbolic
        The index of refraction of the source medium.

    nr2 : float, symbolic
        The index of refraction of the target medium.

    Returns
    -------
    out : complex, :class:`.BeamParam`, symbolic, :class:`numpy.ndarray`
        The transformed beam parameter. If `ABCD` is symbolic then a symbolic
        expression is returned. If `q1` is a :class:`.BeamParam` instance then
        a new :class:`.BeamParam` instance is returned. Note that these two
        conditions couple such that a :class:`.BeamParam` with a symbolic q
        attribute may be returned.

        If `q1` is an array of beam parameters then an array of the above types
        will be returned (dependent upon whether `q1` contains symbolic or numeric
        beam parameters).
    """
    make_bp = isinstance(q1, BeamParam)
    if make_bp:
        q1_symbolic = q1.symbolic

        wl = q1.wavelength
        q1 = q1.q
    else:
        q1_symbolic = isinstance(q1, Symbol)

    # dtype of object for ABCD indicates symbolic ABCD matrix
    if ABCD.dtype == object or q1_symbolic or isinstance(q1, np.ndarray):
        q1_factor = q1 / nr1
        A = ABCD[0][0]
        B = ABCD[0][1]
        C = ABCD[1][0]
        D = ABCD[1][1]

        q2 = nr2 * (A * q1_factor + B) / (C * q1_factor + D)
    # purely numeric so use Cython function for speed
    else:
        q2 = transform_q(ABCD, q1, nr1, nr2)

    if make_bp:
        return BeamParam(q=q2, wavelength=wl, nr=nr2)

    return q2


# TODO (sjr) Cite my thesis derivation of below equation eventually
def ws_overlap(W, S, Wp, Sp, wavelength=None):
    r"""Calculates the WS phase space overlap.

    This overlap function is computed by,

    .. math::

        O(W,S) = \frac{4}{W^2 W_P^2} \frac{1}{\left(\frac{1}{W^2} + \frac{1}{W_p^2}\right)^2 + \frac{k^2}{4}\left(S - S_P\right)^2},

    .. note::
        The equation above was derived from the work in :cite:`PhysRevD.101.102005`.

    where :math:`W` and :math:`S` are the beam size and defocus, respectively, of a gaussian beam propagated
    to the same plane as the primary mode (which has :math:`W_p`, :math:`S_p`).

    Parameters
    ----------
    W : number, array-like
        Beam size of mode [metres].

    S : number, array-like
        Defocus of mode [1/metres].

    Wp : number, array-like
        Beam size of primary mode [metres].

    Sp : number, array-like
        Defocus of primary mode [1/metres].

    wavelength : float, optional
        Wavelength of the beam. Defaults to the value in the
        loaded config file.

    Returns
    -------
    O : number, array-like
        The overlap between the mode :math:`(W,S)` and the primary mode :math:`(W_p,S_p)`.
    """
    if wavelength is None:
        wavelength = config_instance()["constants"].getfloat("lambda0")

    k = 2 * np.pi / wavelength
    A = 4 / (W ** 2 * Wp ** 2)
    B = 1 / W ** 2 + 1 / Wp ** 2
    C = (k / 2) * (S - Sp)

    return A / (B ** 2 + C ** 2)


def ws_overlap_grid(qp, woffset, soffset, wpts=200, spts=200):
    r"""Computes the WS phase space overlap with the primary beam parameter `qp`
    over a grid of W, S data.

    See :func:`.ws_overlap` for a definition of the overlap quantity.

    The :math:`W` and :math:`S` spaces are formed by creating arrays around the
    :attr:`.BeamParam.w` and :attr:`.BeamParam.S` values of `qp`, according to
    the offsets given in `woffset` and `soffset`, respectively.

    Parameters
    ----------
    qp : :class:`.BeamParam`
        The primary "mode" which overlapping modes are calculated against.

    woffset : float, sequence
        A single number, or size 2 sequence. Defines the offsets (lower, upper)
        from the primary mode beamsize to use for the W space.

    soffset : float, sequence
        A single number, or size 2 sequence. Defines the offsets (lower, upper)
        from the primary mode beamsize to use for the S space.

    wpts : int, optional; default: 200
        Number of points for the W space array.

    spts : int, optional; default: 200
        Number of points for the S space array.

    Returns
    -------
    W : :class:`numpy.ndarray`
        The W space (as a 2D grid).

    S : :class:`numpy.ndarray`
        The S space (as a 2D grid).

    OL : :class:`numpy.ndarray`
        The overlap as a function of the WS phase space (as a 2D grid).

    Examples
    --------
    In the following example, we compute the overlap to a primary mode which
    has a 6 mm beam size and a defocus of 0 m (i.e. at the waist).

    .. jupyter-execute::

        import finesse
        finesse.configure(plotting=True)

        import finesse.gaussian as gaussian
        from finesse.plotting.plot import ws_phase_space

        import matplotlib.pyplot as plt

        # Make a beam parameter with w = 6 mm, S = 0 m
        qp = gaussian.BeamParam(w=6e-3, S=0)

        # Compute the WS phase space overlap with qp, using
        # maximum offset of 1 mm in beam size and 1 cm in defocus
        W, S, OL = gaussian.ws_overlap_grid(qp, woffset=1e-3, soffset=1e-2)

        # Now plot this as contours of overlap with custom levels
        fig, ax = ws_phase_space(
            W, S, OL,
            levels=[0.6, 0.8, 0.84, 0.88, 0.92, 0.94, 0.96, 0.98, 0.995, 0.999],
        )
    """
    if not isinstance(qp, BeamParam):
        raise TypeError("Expected argument qp to be of type BeamParam.")

    if qp.symbolic:
        LOGGER.warning(
            "Beam parameter argument 'qp' is symbolic, evaluating this "
            "with current parameter values for use in ws_overlap_grid."
        )
        qp = qp.eval()

    # Get the beam size and defocus of the primary mode
    Wp = qp.w
    Sp = qp.S
    wavelength = qp.wavelength

    if not is_iterable(woffset):
        wlim1 = wlim2 = woffset
    else:
        wlim1, wlim2 = woffset

    if not is_iterable(soffset):
        slim1 = slim2 = soffset
    else:
        slim1, slim2 = soffset

    # Construct the w (beam size), s (defocus) spaces as
    # array offsets around primary beam parameter
    w = np.linspace(Wp - wlim1, Wp + wlim2, wpts)
    s = np.linspace(Sp - slim1, Sp + slim2, spts)

    W, S = np.meshgrid(w, s)
    OL = ws_overlap(W, S, Wp, Sp, wavelength)

    return W, S, OL


class BeamParam:
    r"""Gaussian beam complex parameter.

    This can be a symbolic beam parameter or numeric. For example,
    if :func:`transform_beam_param` is called using a symbolic ABCD
    matrix then a symbolic `BeamParam` instance will be created
    automatically.

    The wavelength of the laser beam defaults to the current value for
    ``lambda0`` in the config file being used, if it is not given as an
    argument during construction of a BeamParam.

    The index of refraction of the associated medium defaults to unity, if
    it is not given as an argument during construction of a BeamParam.

    The following are legal initialisations of a `BeamParam`
    object::

        q = BeamParam(w0=w0, z=z)
        q = BeamParam(z=z, zr=zr)
        q = BeamParam(w=w, rc=rc)
        q = BeamParam(w=w, S=S)
        q = BeamParam(q=c) # where c is a complex number or symbol

    Descriptions of these variables are as follows:

    - :math:`z` --- distance to the waist position (in metres). Negative
      values indicate a beam converging to the waist, positive is a beam
      diverging from the waist.
    - :math:`w_0` --- radius of the beam at the waist (in metres).
    - :math:`zr` --- Rayleigh range of the beam (in metres).
    - :math:`w` --- radius of the beam at the current position (in metres).
    - :math:`rc` --- radius of curvature of the wavefront (in metres).
    - :math:`S` --- curvature (or defocus) of the wavefront (in metres).
    - :math:`q` --- the complex beam parameter itself.

    The default wavelength and refractive index values can also be
    changed with (for example)::

        q = BeamParam(wavelength, nr, w0=w0, zr=zr)

    Note that BeamParam objects can also be symbolic by passing any value
    in `*args` or `**kwargs` as a symbolic expression.

    Parameters
    ----------
    wavelength : float, optional
        Wavelength of the beam light, defaults to the value given for
        ``lambda0`` in the config file being used.

    nr : float, optional
        Refractive index, defaults to unity.
    """

    def __init__(self, wavelength=None, nr=1, *args, **kwargs):
        if wavelength is not None:
            self.__lambda = wavelength
        else:
            self.__lambda = config_instance()["constants"].getfloat("lambda0")
        self.__nr = nr
        self.__symbolic = False

        err_msg = (
            "Expected one of:\n"
            " - q\n"
            " - w0, z\n"
            " - zr, z\n"
            " - w, Rc\n"
            " - w, S\n"
        )

        if len(args) == 1:
            if isinstance(args[0], Symbol):
                self.__q = args[0]
                self.__symbolic = True
            else:
                self.__q = complex(args[0])

        elif len(kwargs) == 1:
            if "q" in kwargs:
                if isinstance(kwargs["q"], Symbol):
                    self.__q = kwargs["q"]
                    self.__symbolic = True
                else:
                    self.__q = complex(kwargs["q"])
            else:
                raise ValueError(err_msg)

        elif len(kwargs) == 2:
            if "w0" in kwargs and "z" in kwargs:
                q = kwargs["z"] + 1j * math.pi * kwargs["w0"] ** 2 / (
                    self.__lambda / self.__nr
                )
            elif "z" in kwargs and "zr" in kwargs:
                q = kwargs["z"] + 1j * kwargs["zr"]
            elif ("rc" in kwargs or "Rc" in kwargs) and "w" in kwargs:
                Rc = kwargs.get("Rc") or kwargs.get("rc")
                one_q = 1 / Rc - 1j * self.__lambda / (
                    math.pi * self.__nr * kwargs["w"] ** 2
                )
                q = 1 / one_q
            elif "S" in kwargs and "w" in kwargs:
                one_q = kwargs["S"] - 1j * self.__lambda / (
                    math.pi * self.__nr * kwargs["w"] ** 2
                )
                q = 1 / one_q
            else:
                raise ValueError(err_msg)

            self.__symbolic = any(isinstance(v, Symbol) for v in kwargs.values())

            self.__q = q
        else:
            raise ValueError(err_msg)

    def __repr__(self):
        return "<%s (w0=%s, w=%s, z=%s) at %s>" % (
            self.__class__.__name__,
            self.w0,
            self.w,
            self.z,
            hex(id(self)),
        )

    @property
    def symbolic(self):
        """Flag indicating whether the beam parameter is a symbolic object.

        :getter: True if symbolic, false otherwise (read-only).
        """
        return self.__symbolic

    def depends_on(self):
        """A list of the model parameters that the symbolic beam parameter depends upon.

        If this beam parameter is not symbolic, then an empty list is returned.
        """
        if not self.symbolic:
            return []

        return self.q.parameters()

    def eval(self, as_bp=True, subs=None):
        """Evaluate the symbolic beam parameter. Parameter substitutions can be
        performed via the substitution dict `subs`.

        Note that this simply returns :attr:`BeamParam.q` if this is not symbolic.

        Parameters
        ----------
        as_bp : bool, optional; default: True
            Flag indicating whether to return result as a :class:`.BeamParam` instance.

        subs : dict, optional; default: None
            Dictionary of parameter substitutions.
        """
        if not self.symbolic:
            q = self.q
        else:
            q = self.q.eval(subs=subs)
            as_bp &= not isinstance(q, np.ndarray)

        if as_bp:
            return BeamParam(wavelength=self.wavelength, q=q, nr=self.nr)

        return q

    @property
    def wavelength(self):
        """The wavelength of the beam (in metres).

        :getter: Returns the wavelength of the beam.
        :setter: Sets the wavelength of the beam.
        """
        return self.__lambda

    @wavelength.setter
    def wavelength(self, value):
        self.__lambda = value

    @property
    def nr(self):
        """The refractive index associated with the `BeamParam`.

        :getter: Returns the index of refraction.
        :setter: Sets the index of refraction.
        """
        return self.__nr

    @nr.setter
    def nr(self, value):
        self.__nr = value

    @property
    def q(self):
        """The complex beam parameter value (:math:`q`).

        :getter: Returns the complex beam parameter value.
        """
        return self.__q

    @property
    def z(self):
        """The relative distance to the waist of the beam (in metres).

        :getter: Returns the relative distance to the waist.
        :setter: Sets the relative distance to the waist.
        """
        if self.__symbolic:
            return self.__q.real()

        return self.__q.real

    @z.setter
    def z(self, value):
        if self.__symbolic:
            self.__q = value + 1j * self.zr
        else:
            self.__q = complex(1j * self.zr + float(value))

    @property
    def zr(self):
        """The Rayleigh range (:math:`z_R`) of the beam (in metres).

        :getter: Returns the Rayleigh range.
        :setter: Sets the Rayleigh range.
        """
        if self.__symbolic:
            return self.__q.imag()

        return self.__q.imag

    @zr.setter
    def zr(self, value):
        if self.__symbolic:
            self.__q = self.z + 1j * value
        else:
            self.__q = complex(self.z + 1j * float(value))

    @property
    def w(self):
        r"""The radius of the beam, :math:`w`, in metres.

        Computed as,

        .. math::

            w = |q|\,\sqrt{\frac{\lambda}{n_r \pi z_R}},

        where :math:`q` is the complex beam parameter (given by :attr:`BeamParam.q`),
        :math:`\lambda` is the beam wavelength (given by :attr:`BeamParam.wavelength`),
        :math:`n_r` is the refractive index of the associated medium (given by
        :attr:`BeamParam.nr`) and :math:`z_R` is the Rayleigh range (given by
        :attr:`BeamParam.zr`, or :attr:`BeamParam.imag`).

        .. hint::

            Use :meth:`.BeamParam.beamsize` to compute the beam size as a function of
            any of the above dependent arguments.

        :getter: Returns the radius, often denoted as "beam-size", of the beam. Read-only.
        """
        return np.abs(self.__q) * np.sqrt(
            self.__lambda / (self.__nr * math.pi * self.zr)
        )

    @property
    def divergence(self):
        r"""Divergence of the beam.

        The divergence is defined as,

        .. math::

            d = \frac{\lambda}{w_0 \pi},

        where :math:`\lambda` is the wavelength and :math:`w_0`
        is the waist size of the beam.

        :getter: Returns the beam divergence. Read-only.
        """
        return self.wavelength / (self.w0 * np.pi)

    @property
    def w0(self):
        r"""The radius of the waist, :math:`w_0`, of the beam, in metres.

        Computed as,

        .. math::

            w_0 = \sqrt{\frac{z_R \lambda}{n_r \pi}},

        where :math:`z_R` is the Rayleigh range (given by :attr:`BeamParam.zr`,
        or :attr:`BeamParam.imag`), :math:`\lambda` is the beam wavelength (given
        by :attr:`BeamParam.wavelength`) and :math:`n_r` is the refractive index of
        the associated medium (given by :attr:`BeamParam.nr`).

        :getter: Returns the beam waist-size.
        :setter: Sets the beam waist-size.
        """
        return np.sqrt(self.zr * self.__lambda / (self.__nr * math.pi))

    @w0.setter
    def w0(self, value):
        if self.__symbolic:
            self.__q = self.z + 1j * value ** 2 * (self.__nr * math.pi) / self.__lambda
        else:
            self.__q = complex(
                self.z + 1j * value ** 2 * (self.__nr * math.pi) / self.__lambda
            )

    @property
    def Rc(self):
        r"""Radius of curvature, :math:`R_c`, of the beam, in metres.

        Computed as,

        .. math::

            R_c = z \left(1 + \left(\frac{z_R}{z}\right)^2\right),

        where :math:`z` is the distance to the waist position (given by
        :attr:`BeamParam.z`, or :attr:`BeamParam.real`) and :math:`z_R` is
        the Rayleigh range (given by :attr:`BeamParam.zr`, or :attr:`BeamParam.imag`).

        Note that if :math:`z = 0` then this will return ``np.inf``.

        .. hint::

            Use :meth:`.BeamParam.roc` to compute the radius of curvature as a
            function of any of the above dependent arguments.

        :getter: Returns the beams' radius of curvature. Read-only.
        """
        if not self.z:
            return np.inf

        return self.z * (1 + (self.zr / self.z) ** 2)

    @property
    def S(self):
        r"""Defocus (wavefront curvature), :math:`S`, of the Gaussian beam, equivalent to
        the reciprocal of :attr:`BeamParam.Rc`.

        Computed as,

        .. math::

            S = \frac{z}{z^2 + z_R^2},

        where :math:`z` is the distance to the waist position (given by
        :attr:`BeamParam.z`, or :attr:`BeamParam.real`) and :math:`z_R` is
        the Rayleigh range (given by :attr:`BeamParam.zr`, or :attr:`BeamParam.imag`).

        .. hint::

            Use :meth:`.BeamParam.curvature` to compute the curvature as a function of
            any of the above dependent arguments.

        :getter: Returns the defocus of the beam parameter. Read-only.
        """
        return self.z / (self.z * self.z + self.zr * self.zr)

    @property
    def psi(self):
        r"""Gouy phase, :math:`\psi`, of the Gaussian beam.

        Computed as,

        .. math::

            \psi = \arctan{\left(\frac{z}{z_R}\right)},

        where :math:`z` is the distance to the waist position (given by
        :attr:`BeamParam.z`, or :attr:`BeamParam.real`) and :math:`z_R` is
        the Rayleigh range (given by :attr:`BeamParam.zr`, or :attr:`BeamParam.imag`).

        .. hint::

            Use :meth:`.BeamParam.gouy` to compute the Gouy phase as a function of
            any of the above dependent arguments.

        :getter: Returns the Gouy phase (in radians) of the beam parameter. Read-only.
        """
        if self.symbolic:
            atan2 = FUNCTIONS["arctan2"]
        else:
            atan2 = np.arctan2

        return atan2(self.z, self.zr)

    def __make_q_from_args(
        self, z=None, wavelength=None, nr=None, w0=None, return_modified_args=False
    ):
        if z is None:
            z = self.z
        else:
            z = np.array(z)

        if wavelength is None:
            wavelength = self.wavelength
        else:
            wavelength = np.array(wavelength)

        if nr is None:
            nr = self.nr
        else:
            nr = np.array(nr)

        if w0 is None:
            w0 = self.w0
        else:
            w0 = np.array(w0)

        q = z + 1j * math.pi * w0 ** 2 / (wavelength / nr)

        if not return_modified_args:
            return q

        return q, z, wavelength, nr, w0

    def beamsize(self, z=None, wavelength=None, nr=None, w0=None):
        """Computes the radius of the beam as a function of any of the dependent
        arguments.

        Calling this method with no arguments is equivalent to
        accessing :attr:`BeamParam.w`.

        Parameters
        ----------
        z : float, optional
            Distance along optical axis relative to waist, defaults
            to :attr:`BeamParam.z` for this beam parameter.

        wavelength : float, optional
            Wavelength of the beam to use, defaults to :attr:`BeamParam.wavelength`
            for this beam parameter.

        nr : float, optional
            Refractive index to use, defaults to :attr:`BeamParam.nr` for
            this beam parameter.

        w0 : float, optional
            Waist-size of the beam, defaults :attr:`BeamParam.w0` for this
            beam parameter.

        Returns
        -------
        The radius of the beam using the specified properties.
        """
        q, _, wavelength, nr, _ = self.__make_q_from_args(
            z, wavelength, nr, w0, return_modified_args=True
        )

        if self.symbolic:
            zr_ = q.imag()
        else:
            zr_ = q.imag

        return np.abs(q) * np.sqrt(wavelength / (nr * math.pi * zr_))

    def gouy(self, z=None, wavelength=None, nr=None, w0=None):
        """Computes the Gouy-phase as a function of any of the dependent arguments.

        Parameters
        ----------
        z : float, optional
            Distance along optical axis relative to waist, defaults
            to :attr:`BeamParam.z` for this beam parameter.

        wavelength : float, optional
            Wavelength of the beam to use, defaults to :attr:`BeamParam.wavelength`
            for this beam parameter.

        nr : float, optional
            Refractive index to use, defaults to :attr:`BeamParam.nr` for
            this beam parameter.

        w0 : float, optional
            Waist-size of the beam, defaults :attr:`BeamParam.w0` for this
            beam parameter.

        Returns
        -------
        The accumulated Gouy phase using the specified properties.
        """
        q = self.__make_q_from_args(z, wavelength, nr, w0)

        if self.symbolic:
            atan2 = FUNCTIONS["arctan2"]
            z_ = q.real()
            zr_ = q.imag()
        else:
            atan2 = np.arctan2
            z_ = q.real
            zr_ = q.imag

        return atan2(z_, zr_)

    def roc(self, z=None, wavelength=None, nr=None, w0=None):
        """Radius of curvature of the beam as a function of any of the dependent
        arguments.

        Calling this method with no arguments is equivalent to
        accessing :attr:`BeamParam.Rc`.

        Parameters
        ----------
        z : float, optional
            Distance along optical axis relative to waist, defaults
            to :attr:`BeamParam.z` for this beam parameter.

        wavelength : float, optional
            Wavelength of the beam to use, defaults to :attr:`BeamParam.wavelength`
            for this beam parameter.

        nr : float, optional
            Refractive index to use, defaults to :attr:`BeamParam.nr` for
            this beam parameter.

        w0 : float, optional
            Waist-size of the beam, defaults :attr:`BeamParam.w0` for this
            beam parameter.

        Returns
        -------
        Radius of curvature of the beam using the specified properties.
        """
        q = self.__make_q_from_args(z, wavelength, nr, w0)

        if self.symbolic:
            z_ = q.real()
            zr_ = q.imag()
        else:
            z_ = q.real
            zr_ = q.imag

        return z_ * (1 + (zr_ / z_) ** 2)

    def curvature(self, z=None, wavelength=None, nr=None, w0=None):
        """Curvature of the beam wavefront as a function of any of the dependent
        arguments.

        Calling this method with no arguments is equivalent to
        accessing :attr:`BeamParam.S`.

        Parameters
        ----------
        z : float, optional
            Distance along optical axis relative to waist, defaults
            to :attr:`BeamParam.z` for this beam parameter.

        wavelength : float, optional
            Wavelength of the beam to use, defaults to :attr:`BeamParam.wavelength`
            for this beam parameter.

        nr : float, optional
            Refractive index to use, defaults to :attr:`BeamParam.nr` for
            this beam parameter.

        w0 : float, optional
            Waist-size of the beam, defaults :attr:`BeamParam.w0` for this
            beam parameter.

        Returns
        -------
        Curvature of the beam using the specified properties.
        """
        q = self.__make_q_from_args(z, wavelength, nr, w0)

        if self.symbolic:
            z_ = q.real()
            zr_ = q.imag()
        else:
            z_ = q.real
            zr_ = q.imag

        return z_ / (z_ * z_ + zr_ * zr_)

    @staticmethod
    def overlap(q1, q2):
        r"""
        Computes the projection from one beam parameter to another to give a measure of the
        overlap between the two beam parameters. The quantity computed is,

        .. math::
            \mathcal{O} = \frac{4|\Im{\{q_1\}}\,\Im{\{q_2\}}|}{|q_1^* - q_2|^2}.

        The return values is :math:`\mathcal{O} \in [0, 1]`, where 0 implies complete
        mode mismatch and 1 indicates full mode matching.

        This function was provided by Paul Fulda and Antonio Perreca, which came originally
        from Chris Mueller.

        Parameters
        ----------
        q1 : :class:`.BeamParam`, complex, array-like
            First beam parameter. Note that this can be a numeric or symbolic beam parameter,
            a complex number or an array of complex values.

        q2 : :class:`.BeamParam`
            Second beam parameter. Note that this can be a numeric or symbolic beam parameter,
            a complex number or an array of complex values.

        Returns
        -------
        overlap : float, array-like, :class:`.Symbol`
            The overlap between `q1` and `q2` as defined above.
        """
        if isinstance(q1, BeamParam):
            q1_q = q1.q
        else:
            q1_q = q1

        if isinstance(q2, BeamParam):
            q2_q = q2.q
        else:
            q2_q = q2

        if isinstance(q1_q, Symbol):
            zr1 = q1_q.imag()
        else:
            zr1 = q1_q.imag
        if isinstance(q2_q, Symbol):
            zr2 = q2_q.imag()
        else:
            zr2 = q2_q.imag

        return np.abs(4 * zr1 * zr2) / np.abs(q1_q.conjugate() - q2_q) ** 2

    @staticmethod
    def mismatch(q1, q2):
        r"""
        Computes the mode mismatch via an alternate form of :math:`1 - \mathcal{O}`, where
        :math:`\mathcal{O}` is the overlap (see :meth:`BeamParam.overlap`).

        This method is less susceptible to floating point errors than simply :math:`1 - \mathcal{O}`
        for very small mismatches. The exact form of the quantity computed is,

        .. math::
            \mathcal{M} = \frac{|q_1 - q_2|^2}{|q_1 - q_2^*|^2}.

        The return value is :math:`\mathcal{M} \in [0, 1]`, where 0 implies full mode matching
        and 1 indicates complete mode mismatch.

        Parameters
        ----------
        q1 : :class:`.BeamParam`
            First beam parameter. Note that this can be a numeric or symbolic beam parameter,
            a complex number or an array of complex values.

        q2 : :class:`.BeamParam`
            Second beam parameter. Note that this can be a numeric or symbolic beam parameter,
            a complex number or an array of complex values.

        Returns
        -------
        mismatch : float, array-like, :class:`.Symbol`
            The mismatch between `q1` and `q2` as defined above.
        """
        if isinstance(q1, BeamParam):
            q1_q = q1.q
        else:
            q1_q = q1

        if isinstance(q2, BeamParam):
            q2_q = q2.q
        else:
            q2_q = q2

        return np.abs(q1_q - q2_q) ** 2 / np.abs(q1_q - q2_q.conjugate()) ** 2

    @staticmethod
    def overlap_contour(q1, M, t):
        """This function returns a set of beam parameters that are mismatched to q1 by
        an overlap M. There are multiple beam parameters that can be X% overlapped with
        one particular q value. This function is parameterised with t from 0 to 2pi,
        which can provide all the possible beam parameters that are M% mismatched.

        q1 - reference beam parameter
        M  - Mismatch factor (1-overlap) [0 -> 1]
        t  - Selection parameter [0 -> 2pi]

        Examples
        --------
        Plots the contours of mismatch for 0.1% and 1% from some initial q value::

            import numpy as np
            import matplotlib.pyplot as plt
            import finesse

            qin = finesse.BeamParam(w0=1e-3, z=20)
            t = np.linspace(0, 2*np.pi, 100)

            # use vectorised functions to select a cerain property of the beam paramters
            vx  = np.vectorize(lambda q: q.z)
            vy  = np.vectorize(lambda q: q.w/1e-3)

            for mm in [1e-3, 2e-2]:
                mmc = finesse.BeamParam.overlap_contour(qin, mm, t)

                plt.text(vx(mmc[20]), vy(mmc[20]), "%1.1f%%" % ((mm*100)),alpha=0.5, fontsize=8)
                l, = plt.plot(vx(mmc),     vy(mmc),     ls='--', alpha=0.2, zorder=-10, c='k')

            plt.show()
        """
        from numpy import vectorize

        assert M < 1 and M >= 0
        assert q1.imag > 0

        vbp = vectorize(lambda x: BeamParam(q1.wavelength, q1.nr, x))

        z1 = np.real(q1)
        zR1 = np.imag(q1)
        r = (2 * np.sqrt(M) * zR1) / (1 - M)
        y0 = ((M + 1) * zR1) / (1 - M)
        x0 = z1
        q2 = r * np.cos(t) + x0 + 1j * (r * np.sin(t) + y0)

        return vbp(q2)

    def conjugate(self):
        """Computes and returns the complex conjugate of the beam parameter.

        Returns
        -------
        q_conj : :class:`.BeamParam`
            The complex conjugate of this `BeamParam` instance.
        """
        return BeamParam(self.wavelength, self.nr, self.q.conjugate())

    def __abs__(self):
        return abs(self.q)

    def __complex__(self):
        return self.q

    def __str__(self):
        return (
            f"BeamParam: q = {self.q}\n"
            f"    w0 = {self.w0} m, z = {self.z} m\n"
            f"     w = {self.w} m, Rc = {self.Rc} m"
        )

    def __mul__(self, a):
        return BeamParam(self.wavelength, self.nr, q=self.__q * a)

    def __imul__(self, a):
        self.__q *= a
        return self

    __rmul__ = __mul__

    def __add__(self, a):
        return BeamParam(self.wavelength, self.nr, self.q + a)

    def __iadd__(self, a):
        self.__q += complex(a)
        return self

    __radd__ = __add__

    def __sub__(self, a):
        return BeamParam(self.wavelength, self.nr, self.q - a)

    def __isub__(self, a):
        self.__q -= a
        return self

    def __rsub__(self, a):
        return BeamParam(self.wavelength, self.nr, a - self.q)

    def __div__(self, a):
        return BeamParam(self.wavelength, self.nr, self.q / a)

    def __truediv__(self, a):
        if isinstance(a, BeamParam):
            return BeamParam(self.wavelength, self.nr, self.q / a.q)

        return BeamParam(self.wavelength, self.nr, self.q / a)

    def __idiv__(self, a):
        self.__q /= a
        return self

    def __pow__(self, q):
        return BeamParam(self.wavelength, self.nr, self.q ** q)

    def __neg__(self):
        return BeamParam(self.wavelength, self.nr, -self.q)

    def __eq__(self, q):
        if q is None:
            return False

        if self.symbolic:
            if isinstance(q, Symbol) or (isinstance(q, BeamParam) and q.symbolic):
                return np.allclose(self.eval(), q.eval())

            return ceq(complex(q), self.eval())

        return ceq(complex(q), self.q)

    @property
    def real(self):
        """The real part of the complex beam parameter, equal to the relative distance
        to the beam waist (in metres).

        :getter: Returns the real part of the beam parameter.
        :setter: Sets the real part of the beam parameter.
        """
        return self.z

    @real.setter
    def real(self, value):
        self.z = value

    @property
    def imag(self):
        """The imaginary part of the complex beam parameter, equal to the Rayleigh range
        :math:`z_R` of the beam (in metres).

        :getter: Returns the imaginary part of the beam parameter.
        :setter: Sets the imaginary part of the beam parameter.
        """
        return self.zr

    @imag.setter
    def imag(self, value):
        self.zr = value

    def reverse(self):
        """Returns the reversed beam parameter.

        This is the beam parameter with the sign of the distance to waist flipped,
        i.e. :math:`q_r = -q^*` where :math:`q_r` is the reversed beam parameter.

        Returns
        -------
        q_r : :class:`BeamParam`
            The reverse of this beam parameter.
        """
        return BeamParam(self.wavelength, self.nr, -1.0 * self.q.conjugate())


class HGMode:
    """An object representation of a Hermite-Gauss mode.

    Parameters
    ----------
    q : :class:`.BeamParam` or complex, or length two sequence of
        The beam parameter. Specify a tuple of (qx, qy) as this argument
        for an astigmatic beam.

    n : int, optional; default: 0
        Tangential mode index.

    m : int, optional; default: 0
        Sagittal mode index.
    """

    def __init__(self, q, n=0, m=0):
        if is_iterable(q):
            qx, qy = q
        else:
            qx = qy = q

        if not isinstance(qx, BeamParam):
            qx = BeamParam(q=qx)
        if not isinstance(qy, BeamParam):
            qy = BeamParam(q=qy)

        self.__qx = qx
        self.__qy = qy

        if self.qx.nr != self.qy.nr:
            raise ValueError("Refractive indices associated with qs must be equal.")
        if self.qx.wavelength != self.qy.wavelength:
            raise ValueError("Wavelengths associated with qs must be equal.")

        nr = self.qx.nr
        lambda0 = self.qx.wavelength
        self._workspace = HGModeWorkspace(
            int(n), int(m), self.qx.q, self.qy.q, nr, lambda0
        )

    @property
    def n(self):
        return self._workspace.n

    @property
    def m(self):
        return self._workspace.m

    @property
    def qx(self):
        return self.__qx

    @qx.setter
    def qx(self, value):
        if not isinstance(value, BeamParam):
            value = BeamParam(q=value)

        self.__qx = value
        self._workspace.set_values(
            qx=self.qx, nr=self.qx.nr, lambda0=self.qx.wavelength
        )

    @property
    def qy(self):
        return self.__qy

    @qy.setter
    def qy(self, value):
        if not isinstance(value, BeamParam):
            value = BeamParam(q=value)

        self.__qy = value
        self._workspace.set_values(
            qy=self.qy, nr=self.qy.nr, lambda0=self.qy.wavelength
        )

    def set_q(self, q):
        """Sets the beam parameter in both planes to the same value `q`.

        Parameters
        ----------
        q : :class:`.BeamParam` or complex
            The beam parameter.
        """
        if not isinstance(q, BeamParam):
            q = BeamParam(q=q)

        self.__qx = q
        self.__qy = q

        nr = q.nr
        lambda0 = q.wavelength
        self._workspace.set_values(qx=self.qx, qy=self.qy, nr=nr, lambda0=lambda0)

    def __check_shapes(self, a, out, direction):
        if out.shape != a.shape:
            raise ValueError(
                f"Shape of output array ({out.shape}) not equal to "
                f"shape of {direction} array ({a.shape})"
            )

    def un(self, x, out=None):
        """Compute the beam profile in the tangential plane."""

        if out is not None:
            self.__check_shapes(x, out, "x")

        return self._workspace.u_n(x, out=out)

    def um(self, y, out=None):
        """Compute the beam profile in the tangential plane."""

        if out is not None:
            self.__check_shapes(y, out, "y")

        return self._workspace.u_m(y, out=out)

    def unm(self, x, y, out=None):
        """Compute the full transverse beam profile."""

        return self._workspace.u_nm(x, y, out=out)



from finesse.detectors.compute.camera import (
    field_pixel_output,
    ccd_pixel_output,
    field_line_output,
    ccd_line_output,
    field_camera_output,
    ccd_output,
)
from finesse.detectors.compute.power import (
    pd0_DC_output,
    pd0_DC_output_segmented,
    pd0_DC_output_masked,
    pd1_DC_output,
    pd1_AC_output,
    pd2_DC_output,
    pd2_AC_output,
)


__all__ = (
    "field_pixel_output",
    "ccd_pixel_output",
    "field_line_output",
    "ccd_line_output",
    "field_camera_output",
    "ccd_output",
    "pd0_DC_output",
    "pd0_DC_output_masked",
    "pd1_DC_output",
    "pd1_AC_output",
    "pd2_DC_output",
    "pd2_AC_output",
)



"""
Optical components performing directional redirection of beams.
"""

import logging
import numpy as np

from finesse.components.general import Connector, InteractionType
from finesse.components.node import NodeDirection, NodeType
from finesse.components.workspace import Connections
from finesse.components.modal.workspace import KnmConnectorWorkspace
from finesse.utilities import refractive_index

LOGGER = logging.getLogger(__name__)


class DBSWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, sim, refill):
        super().__init__(owner, sim, refill, refill)


class DirectionalBeamsplitter(Connector):
    """
    Represents a directional beamsplitter optical component.

    Parameters
    ----------
    name : str
        Name of newly created directional beamsplitter.
    """

    def __init__(self, name):
        super().__init__(name)

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p3", NodeType.OPTICAL)
        self.p3._add_node("i", NodeDirection.INPUT)
        self.p3._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p4", NodeType.OPTICAL)
        self.p4._add_node("i", NodeDirection.INPUT)
        self.p4._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P3o",
            self.p1.i,
            self.p3.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P3i_P4o",
            self.p3.i,
            self.p4.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P4i_P2o",
            self.p4.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )

    def _get_workspace(self, sim):
        ws = DBSWorkspace(self, sim, False)
        ws.I = np.eye(sim.model_data.num_HOMs, dtype=np.complex128)
        ws.carrier.add_fill_function(self._fill_carrier, False)
        if sim.signal:
            ws.signal.add_fill_function(self._fill_signal, False)

        ws.nr1 = refractive_index(self.p1)
        ws.nr2 = refractive_index(self.p2)
        ws.nr3 = refractive_index(self.p3)
        ws.nr4 = refractive_index(self.p4)

        if sim.is_modal:
            ws.set_knm_info(
                "P1i_P3o", nr_from=ws.nr1, nr_to=ws.nr3, is_transmission=True
            )
            ws.set_knm_info(
                "P3i_P4o", nr_from=ws.nr3, nr_to=ws.nr4, is_transmission=True
            )
            ws.set_knm_info(
                "P4i_P2o", nr_from=ws.nr4, nr_to=ws.nr2, is_transmission=True
            )
            ws.set_knm_info(
                "P2i_P1o", nr_from=ws.nr2, nr_to=ws.nr1, is_transmission=True
            )
        return ws

    def _fill_carrier(self, ws):
        self._fill_matrix(ws.owner_id, ws.sim.carrier, ws.carrier.connections, ws.I)

    def _fill_signal(self, ws):
        self._fill_matrix(ws.owner_id, ws.sim.signal, ws.signal.connections, ws.I)

    def _fill_matrix(self, owner_id, mtx, connections, I):
        for freq in mtx.optical_frequencies.frequencies:
            for idx in (
                connections.P1i_P3o_idx,
                connections.P3i_P4o_idx,
                connections.P4i_P2o_idx,
                connections.P2i_P1o_idx,
            ):
                with mtx.component_edge_fill3(
                    owner_id, idx, freq.index, freq.index,
                ) as mat:
                    mat[:] = I



"""Graph plotting."""

from collections import defaultdict
import matplotlib.pyplot as plt
import networkx as nx
from ..utilities import graph_layouts, option_list


def plot_graph(
    network, layout, graphviz=False, **kwargs,
):
    from ..environment import has_pygraphviz

    if graphviz and not has_pygraphviz():
        raise ModuleNotFoundError(
            "The graphviz option requires pygraphviz and graphviz to be installed"
        )

    plotter = plot_graphviz if graphviz else plot_nx_graph
    plotter(network, layout, **kwargs)


def plot_nx_graph(
    network,
    layout,
    node_labels=True,
    node_attrs=False,
    edge_attrs=False,
    node_color_key=None,
    label_font_size=12,
    attr_font_size=6,
    edge_font_size=6,
    **kwargs,
):
    """Plot graph with NetworkX.

    Parameters
    ----------
    network : :class:`networkx.Graph`
        The network to plot.

    layout : str
        The layout type to use. Any layout algorithm provided by
        :mod:`networkx.drawing.layout` is supported.

    node_labels : :py:class:`bool`, optional
        Show node names; defaults to True.

    node_attrs : :py:class:`bool` or :py:class:`list`, optional
        Show node data. This can be `True`, in which case all node data is shown, or a
        list, in which case only the specified keys are shown. Defaults to `True`.

    edge_attrs : :py:class:`bool` or :py:class:`list`, optional
        Show edge data. This can be `True`, in which case all edge data is shown, or a list, in
        which case only the specified keys are shown. Defaults to `True`.

    node_color_key : callable, optional
        Key function accepting a node and a node attribute dict and returning a group. Each group is
        assigned a unique color. If not specified, nodes are not colored.

    label_font_size, attr_font_size, edge_font_size : :py:class:`int`, optional
        Font size for node labels, attributes and edges. Defaults to 12, 6 and 6, respectively.

    Other Parameters
    ----------------
    kwargs
        Anything else supported by :func:`networkx.drawing.nx_pylab.draw`.

    Raises
    ------
    ValueError
        If the specified layout is not supported.

    Exception
        If the graph cannot be represented with the specified layout.
    """
    from ..utilities import stringify

    if node_color_key is not None:
        if "node_color" in kwargs:
            raise ValueError(
                "cannot specify both 'node_color' and 'node_color_key' arguments"
            )

        # Assign node colors.
        cycler = iter(plt.rcParams["axes.prop_cycle"].by_key()["color"])
        group_colors = defaultdict(lambda: next(cycler))
        kwargs["node_color"] = [
            group_colors[node_color_key(node, data)]
            for node, data in network.nodes(data=True)
        ]

    layouts = graph_layouts()

    try:
        posfunc = layouts[layout.casefold()]
    except KeyError:
        choices = option_list(layouts)

        raise ValueError(
            f"Layout '{layout}' is not available in NetworkX (choose from {choices})."
        )

    try:
        pos = posfunc(network)
    except nx.NetworkXException as e:
        if "G is not planar" in str(e):
            raise Exception(
                "Graph cannot be represented with a planar layout. Try a different layout."
            ) from e

    nx.draw(
        network,
        pos,
        with_labels=node_labels,
        verticalalignment="bottom",
        font_size=label_font_size,
        **kwargs,
    )

    if node_attrs:
        data = network.nodes(data=True)

        if node_attrs is not True:  # Needs to be like this!
            # Show only certain data.
            data = [
                (
                    node,
                    {
                        key: value
                        for key, value in node_data.items()
                        if key in node_attrs
                    },
                )
                for node, node_data in data
            ]

        node_labels = {
            node: "\n".join(
                [f"{key}={stringify(value)}" for key, value in node_attrs.items()]
            )
            for node, node_attrs in data
        }
        nx.draw_networkx_labels(
            network,
            pos,
            labels=node_labels,
            verticalalignment="top",
            font_size=attr_font_size,
        )

    if edge_attrs:
        data = network.edges(data=True)

        if edge_attrs is not True:  # Needs to be like this!
            # Show only certain data.
            data = (
                (
                    u,
                    v,
                    {
                        key: value
                        for key, value in edge_data.items()
                        if key in edge_attrs
                    },
                )
                for u, v, edge_data in data
            )

        edge_labels = {
            (u, v): "\n".join(
                [f"{key}={stringify(value)}" for key, value in edge_attrs.items()]
            )
            for u, v, edge_attrs in data
        }
        nx.draw_networkx_edge_labels(
            network, pos, edge_labels=edge_labels, font_size=edge_font_size,
        )

    plt.show()


def plot_graphviz(network, layout):
    """Plot graph with graphviz.

    The `pygraphviz` Python package must be installed and available on the current
    Python path, and `graphviz` must be available on the system path.

    Parameters
    ----------
    network : :class:`networkx.Graph`
        The network to plot.

    layout : str
        The layout type to use. Any layout algorithm provided by graphviz is supported.

    Raises
    ------
    ValueError
        If the specified layout is not supported.

    ImportError
        If graphviz or pygraphviz is not installed.
    """
    from networkx.drawing.nx_agraph import view_pygraphviz

    layouts = ("neato", "dot", "fdp", "sfdp", "circo")
    gvlayout = layout.casefold()

    if gvlayout not in layouts:
        choices = option_list(layouts)

        raise ValueError(
            f"Layout '{layout}' is not available in graphviz (choose from {choices})."
        )

    view_pygraphviz(network, prog=gvlayout)



"""
Signal-type electrical component for producing signal inputs.
"""
from finesse.components.node import NodeType, Node, Port
from finesse.components.general import Connector
from finesse.parameter import float_parameter, info_parameter
from finesse.components.modal.signal import siggen_fill_rhs, SignalGeneratorWorkspace
from finesse.components.dof import DegreeOfFreedom

@float_parameter("amplitude", "Amplitude", units="arb")
@float_parameter("phase", "Phase", units="degrees")
@info_parameter("f", "Frequency", units="Hz")
class SignalGenerator(Connector):
    """Represents a signal generator which produces a signal with a given amplitude and phase.

    Parameters
    ----------
    name : str
        Name of newly created signal generator.

    node : .class:`finesse.components.node.Node`
        A node to inject a signal into.

    amplitude : float, optional
        Amplitude of the signal in volts.

    phase : float, optional
        Phase-offset of the signal from the default in degrees, defaults to zero.
    """

    def __init__(self, name, node, amplitude=1, phase=0):
        Connector.__init__(self, name)

        if isinstance(node, DegreeOfFreedom):
            node = node.AC.i
        elif isinstance(node, Port):
            if len(node.nodes) == 1:
                node = node.nodes[0]
            else:
                raise Exception(
                    f"Signal generator ({name}) input `{node}` should be a Node not a Port"
                )
        elif not isinstance(node, Node):
            raise Exception(
                f"Signal generator ({name}) input `{node}` should be a Node"
            )

        self._add_port("port", node.type)
        self.port._add_node("o", None, node)
        node.has_signal_injection = True

        self.amplitude = amplitude
        self.phase = phase

    @property
    def node(self):
        """The node the signal generator injects into.

        :getter: Returns the node the signal generator injects into.
        """
        return self.port.o

    @property
    def f(self):
        return self._model.fsig.f

    def _get_workspace(self, sim):
        if sim.signal:
            ws = SignalGeneratorWorkspace(self, sim, True)
            ws.rhs_idx = ws.sim.signal.field(self.port.o, 0, 0)
            ws.signal.set_fill_rhs_fn(siggen_fill_rhs)

            if self.port.o.type is NodeType.MECHANICAL:
                ws.scaling = 1 / sim.model_data.x_scale
            else:
                ws.scaling = 1

            return ws
        else:
            return None



"""Detectors for capturing images, slices and single pixels of a beam.

The camera types are split into two categories (CCDs and ComplexCameras) based on the
mathematical implementation shown in :ref:`camera_equations`.
"""

from abc import ABC
import numbers

import numpy as np
import logging

from finesse.parameter import float_parameter

from finesse.detectors.general import MaskedDetector
from finesse.detectors.compute.camera import (
    CameraWorkspace,
    CCDWorkspace,
    CCDLineWorkspace,
    FieldCameraWorkspace,
    FieldLineWorkspace,
    FieldPixelWorkspace,
)
from finesse.detectors.compute import (
    field_pixel_output,
    ccd_pixel_output,
    field_line_output,
    ccd_line_output,
    field_camera_output,
    ccd_output,
)
from finesse.utilities.misc import find_nearest, is_iterable


LOGGER = logging.getLogger(__name__)


class Camera(MaskedDetector, ABC):
    """Base camera class.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    w0_scaled : bool
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.

    dtype : :class:`numpy.dtype` or str
        The data-type of the pixels.

    shape : tuple
        Dimensions of the camera image.
    """

    def __init__(self, name, node, w0_scaled, dtype, shape, **kwargs):
        MaskedDetector.__init__(self, name, node, dtype=dtype, shape=shape, **kwargs)
        self.__w0_scaled = w0_scaled in (True, 1, "True", "true", "Y", "y")
        self._nr = self.node.space.nr if self.node.space is not None else 1.0
        self._changing_check = set()

    @property
    def needs_trace(self):
        return True

    @property
    def w0_scaled(self):
        """Flag for whether the x and y co-ordinates have been scaled by the waist-size
        of the beam parameter at the detection node.

        :getter: Returns `True` if x and y have been scaled by the beam waist, `False` otherwise.
        """
        return self.__w0_scaled

    @property
    def scaled_xdata(self):
        """Coordinate data in the x-axis scaled to metres.

        Equivalent to ``xdata`` if :attr:`.Camera.w0_scaled` is False. Otherwise this is
        ``xdata`` multiplied by the tangential waist size as measured at the node.
        """
        x = self.xdata
        if not self.w0_scaled:
            return x

        qx = self._model.beam_trace()[self.node].qx
        return x * qx.w0

    @property
    def scaled_ydata(self):
        """Coordinate data in the y-axis scaled to metres.

        Equivalent to ``ydata`` if :attr:`.Camera.w0_scaled` is False. Otherwise this is
        ``ydata`` multiplied by the sagittal waist size as measured at the node.
        """
        y = self.ydata
        if not self.w0_scaled:
            return y

        qy = self._model.beam_trace()[self.node].qy
        return y * qy.w0


class Image:
    r"""Data structure representation of an image.

    Parameters
    ----------
    xlim : sequence or scalar
        Limits of the x-dimension of the image. If a single number is given then
        this will be computed as :math:`x_{\mathrm{lim}} = [-|x|, +|x|]`.

    ylim : sequence or scalar
        Limits of the y-dimension of the image. If a single number is given then
        this will be computed as :math:`y_{\mathrm{lim}} = [-|y|, +|y|]`.

    npts : int
        Number of points for both the x and y axes.

    dtype : str or dtype
        Data type of the image to pass to NumPy for array creation.
    """

    def __init__(self, xlim, ylim, npts, dtype):
        xl, xu = _check_limits(xlim)
        yl, yu = _check_limits(ylim)

        self.__set_x_space(npts, lower=xl, upper=xu)
        self.__set_y_space(npts, lower=yl, upper=yu)

        self.__set_out_grid(dtype)

    def __set_x_space(self, npts, lower=None, upper=None):
        npts = int(npts)
        if not npts > 0:
            raise ValueError("Number of points must be a positive integer.")

        if lower is None:
            lower = self._x[0]
        if upper is None:
            upper = self._x[-1]

        self._x = np.linspace(lower, upper, npts, dtype=np.float64)

    def __set_y_space(self, npts, lower=None, upper=None):
        npts = int(npts)
        if not npts > 0:
            raise ValueError("Number of points must be a positive integer.")

        if lower is None:
            lower = self._y[0]
        if upper is None:
            upper = self._y[-1]

        self._y = np.linspace(lower, upper, npts, dtype=np.float64)

    def __set_out_grid(self, dtype=None):
        if dtype is None:
            dtype = self._out.dtype

        self._out = np.zeros(self.resolution, dtype=dtype)

    @property
    def xlim(self):
        """The limits of the x coordinate data.

        :getter: Returns a tuple of ``(xmin, xmax)``.
        :setter: Sets the x-axis limits.
        """
        return self._x[0], self._x[-1]

    @xlim.setter
    def xlim(self, value):
        xl, xu = _check_limits(value)
        self.__set_x_space(self.npts, xl, xu)

    @property
    def ylim(self):
        """The limits of the y coordinate data.

        :getter: Returns a tuple of ``(ymin, ymax)``.
        :setter: Sets the y-axis limits.
        """
        return self._y[0], self._y[-1]

    @ylim.setter
    def ylim(self, value):
        yl, yu = _check_limits(value)
        self.__set_y_space(self.npts, yl, yu)

    @property
    def xdata(self):
        """The array of data points for the x-axis.

        :getter: Returns a copy of the :class:`numpy.ndarray` containing the x-axis points.
        """
        return self._x.copy()

    @property
    def ydata(self):
        """The array of data points for the y-axis.

        :getter: Returns a copy of the :class:`numpy.ndarray` containing the y-axis points.
        """
        return self._y.copy()

    @property
    def npts(self):
        """Number of pixels in each axis.

        :getter: Returns the number of pixels in each axis.
        :setter: Sets the number of pixels in each axis.
        """
        return self._x.shape[0]

    @npts.setter
    def npts(self, value):
        self.__set_x_space(value)
        self.__set_y_space(value)
        self.__set_out_grid()

    @property
    def resolution(self):
        """The resolution of the image.

        Currently this is always square (i.e. number of points in both axes
        always equal).

        :getter: Returns the tuple ``(xpts, ypts)``.
        """
        return self.npts, self.npts

    def at(self, x=None, y=None):
        """Retrieves a slice or single pixel of the output image.

        Parameters
        ----------
        x : scalar, optional
            Value indicating where to take a y-slice of the image or,
            if used in conjunction with `y`, which pixel to return. Defaults
            to `None`.

        y : scalar, optional
            Value indicating where to take a x-slice of the image or,
            if used in conjunction with `x`, which pixel to return. Defaults
            to `None`.

        magnitude : bool, optional
            Returns the amplitude of the detected field if `True`. Otherwise
            returns the full complex description.

        Returns
        -------
        out : :class:`numpy.ndarray` or float
            Either a slice of the image or a single pixel at the specified co-ordinates.
        """
        if x is None and y is None:
            return self._out.copy()

        if x is None:
            nearest_idx = find_nearest(self._y, y, index=True)
            values = self._out[nearest_idx][:]
            return values.copy()

        if y is None:
            nearest_idx = find_nearest(self._x, x, index=True)
            values = self._out[:, nearest_idx]
            return values.copy()

        nearest_indices = (
            find_nearest(self._x, x, index=True),
            find_nearest(self._y, y, index=True),
        )
        values = self._out[nearest_indices]
        return values.copy()


class ScanLine:
    r"""Data structure representation of a slice of an image.

    Parameters
    ----------
    x : scalar or None
        The x coordinate of the slice.

    y : scalar or None
        The y coordinate of the slice.

    xlim : scalar or size two sequence
        The limits of the x-axis scan lines. A single number gives
        :math:`x_{\mathrm{axis}} \in [-|x|, +|x|]`, or a tuple of size two gives
        :math:`x_{\mathrm{axis}} \in [x[0], x[1]]`.

    ylim : scalar or array-like
        The limits of the y-axis scan lines. A single number gives
        :math:`y_{\mathrm{axis}} \in [-|y|, +|y|]`, or a tuple of size two gives
        :math:`y_{\mathrm{axis}} \in [y[0], y[1]]`.

    npts : int
        Number of points in slice axis.

    dtype : str or dtype
        Data type of the slice to pass to NumPy for array creation.
    """

    def __init__(self, npts, dtype, x=None, y=None, xlim=None, ylim=None):
        if xlim is not None and ylim is not None:
            raise ValueError("Both xlim and ylim cannot be specified.")

        if xlim is not None:
            self.__direction = "x"
        elif ylim is not None:
            self.__direction = "y"
        else:
            raise ValueError("One of xlim or ylim must be specified.")

        if self.direction == "x":
            xl, xu = _check_limits(xlim)
            self._x = np.linspace(xl, xu, npts, dtype=np.float64)

            if x is not None:
                LOGGER.warning(
                    "Ignoring x = %f argument passed to ScanLine as xlim "
                    "has been specified.",
                    x,
                )

            if y is None:
                self._y = 0
            else:
                self._y = y
        else:
            yl, yu = _check_limits(ylim)
            self._y = np.linspace(yl, yu, npts, dtype=np.float64)

            if y is not None:
                LOGGER.warning(
                    "Ignoring y = %f argument passed to ScanLine as ylim "
                    "has been specified.",
                    y,
                )

            if x is None:
                self._x = 0
            else:
                self._x = x

        self._out = np.zeros(npts, dtype=dtype)

    def __set_ax_space(self, npts, lower=None, upper=None):
        npts = int(npts)
        if not npts > 0:
            raise ValueError("Number of points must be a positive integer.")

        if self.direction == "x":
            ax = self._x
        else:
            ax = self._y

        if lower is None:
            lower = ax[0]
        if upper is None:
            upper = ax[-1]

        ax = np.linspace(lower, upper, npts, dtype=np.float64)
        if self.direction == "x":
            self._x = ax
        else:
            self._y = ax

    def __set_out_line(self, dtype=None):
        if dtype is None:
            dtype = self._out.dtype

        self._out = np.zeros(self.npts, dtype=dtype)

    @property
    def direction(self):
        """The slice axis - i.e. 'x' for x-axis, 'y' for y-axis.

        :getter: Returns a string determining the slice axis (read-only).
        """
        return self.__direction

    @property
    def x(self):
        """The x co-ordinate of the slice.

        If :attr:`.ScanLine.direction` is 'x' then this will return ``None``.

        :getter: Returns the x coordinate of the slice.
        :setter: Sets the x coordinate of the slice.
        """
        if self.direction == "x":
            return None

        return self._x

    @x.setter
    def x(self, value):
        if self.direction == "x":
            raise RuntimeError(
                "Cannot set x-position of scan line when the slice is in the x-axis."
            )

        self._x = float(value)

    @property
    def y(self):
        """The y co-ordinate of the slice.

        If :attr:`ScanLine.direction` is 'y' then this will return ``None``.

        :getter: Returns the y coordinate of the slice.
        :setter: Sets the y coordinate of the slice.
        """
        if self.direction == "y":
            return None

        return self._y

    @y.setter
    def y(self, value):
        if self.direction == "y":
            raise RuntimeError(
                "Cannot set y-position of scan line when the slice is in the y-axis."
            )

        self._y = float(value)

    @property
    def xlim(self):
        """The limits of the slice in the x-axis.

        If :attr:`ScanLine.direction` is 'y' then this will return ``None``.

        :getter: Returns a tuple of ``(xmin, xmax)``.
        :setter: Sets the x-axis limits.
        """
        if self.direction == "y":
            return None

        return self._x[0], self._x[-1]

    @xlim.setter
    def xlim(self, value):
        if self.direction == "y":
            raise RuntimeError("Cannot set x-axis limits for scan line in the y-axis.")

        xl, xu = _check_limits(value)
        self.__set_ax_space(self.npts, xl, xu)

    @property
    def ylim(self):
        """The limits of the slice in the y-axis.

        If :attr:`ScanLine.direction` is 'x' then this will return ``None``.

        :getter: Returns a tuple of ``(ymin, ymax)``.
        :setter: Sets the y-axis limits.
        """
        if self.direction == "x":
            return None

        return self._y[0], self._y[-1]

    @ylim.setter
    def ylim(self, value):
        if self.direction == "x":
            raise RuntimeError("Cannot set y-axis limits for scan line in the x-axis.")

        yl, yu = _check_limits(value)
        self.__set_ax_space(self.npts, yl, yu)

    @property
    def xdata(self):
        """The numeric value(s) of the x coordinate.

        If :attr:`ScanLine.direction` is 'x' then this will be a copy of the array of
        values, otherwise it is a single value equivalent to :attr:`.ScanLine.x`.

        :getter: The x coordinate value(s). Read-only.
        """
        if self.direction == "x":
            return self._x.copy()

        return self._x

    @property
    def ydata(self):
        """The numeric value(s) of the y coordinate.

        If :attr:`ScanLine.direction` is 'y' then this will be a copy of the array of
        values, otherwise it is a single value equivalent to :attr:`.ScanLine.y`.

        :getter: The y coordinate value(s). Read-only.
        """
        if self.direction == "y":
            return self._y.copy()

        return self._y

    @property
    def npts(self):
        """Number of pixels in the scanning axis.

        :getter: Returns the number of pixels in the slice axis.
        :setter: Sets the number of pixels in the slice axis.
        """
        if self.direction == "x":
            return self._x.shape[0]

        return self._y.shape[0]

    @npts.setter
    def npts(self, value):
        self.__set_ax_space(value)
        self.__set_out_line()


class Pixel:
    r"""Data structure representation of a pixel of an image.

    Parameters
    ----------
    x : scalar
        The x co-ordinate of the pixel.

    y : scalar
        The y co-ordinate of the pixel.

    dtype : str or dtype
        Data type of the pixel.
    """

    def __init__(self, x, y, dtype):
        self.x = x
        self.y = y

        if dtype == np.complex128:
            self._out = complex(0, 0)
        else:
            self._out = 0.0

    @property
    def x(self):
        """The x coordinate of the pixel.

        :getter: Returns the x coordinate of the pixel.
        :setter: Sets the x coordinate of the pixel.
        """
        return self._x

    @x.setter
    def x(self, value):
        self._x = float(value)

    @property
    def xdata(self):
        """Equivalent to :attr:`.Pixel.x`.

        :getter: Returns the x coordinate of the pixel. Read-only version.
        """
        return self._x

    @property
    def y(self):
        """The y coordinate of the pixel.

        :getter: Returns the y coordinate of the pixel.
        :setter: Sets the y coordinate of the pixel.
        """
        return self._y

    @y.setter
    def y(self, value):
        self._y = float(value)

    @property
    def ydata(self):
        """Equivalent to :attr:`.Pixel.y`.

        :getter: Returns the y coordinate of the pixel. Read-only version.
        """
        return self._y


class CCDCamera(Camera, ABC):
    """Abstract type for cameras which detect pixel intensity.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, w0_scaled=True, **kwargs):
        if isinstance(self, Image):
            shape = self._x.shape[0], self._y.shape[0]
        elif isinstance(self, ScanLine):
            if self.direction == "x":
                shape = self._x.shape
            else:
                shape = self._y.shape
        elif isinstance(self, Pixel):
            shape = None
        else:
            raise TypeError(
                "Bug detected! CCDCamera does not derive from Image, "
                "ScanLine or Pixel."
            )

        Camera.__init__(self, name, node, w0_scaled, np.float64, shape, **kwargs)


class CCD(CCDCamera, Image):
    r"""Camera for detecting the full image of the beam in terms of the intensity.

    Get the unscaled x and y coordinate data via :attr:`.Image.xdata` and :attr:`.Image.ydata`,
    respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    xlim : sequence or scalar
        Limits of the x-dimension of the image. If a single number is given then
        this will be computed as :math:`x_{\mathrm{lim}} = [-|x|, +|x|]`.

    ylim : sequence or scalar
        Limits of the y-dimension of the image. If a single number is given then
        this will be computed as :math:`y_{\mathrm{lim}} = [-|y|, +|y|]`.

    npts : int
        Number of points in both axes.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, xlim, ylim, npts, w0_scaled=True):
        Image.__init__(self, xlim, ylim, npts, dtype=np.float64)
        CCDCamera.__init__(self, name, node, w0_scaled)

    @property
    def npts(self):
        return super().npts

    @npts.setter
    def npts(self, value):
        super(CCD, self.__class__).npts.fset(self, value)
        self._update_dtype_shape(self.resolution)

    def _get_workspace(self, sim):
        ws = CCDWorkspace(self, sim, self._out)
        ws.set_output_fn(ccd_output)
        return ws


class CCDScanLine(CCDCamera, ScanLine):
    r"""Camera for detecting a slice of the beam in terms of the intensity.

    The :attr:`.ScanLine.direction` (i.e. axis of slice) is determined
    from which of `xlim` or `ylim` is specified.

    Get the unscaled x and y coordinate data via :attr:`.ScanLine.xdata` and
    :attr:`.ScanLine.ydata`, respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    npts : int
        Number of points in slice axis.

    x : scalar or None; default: None
        The x coordinate of the slice. If ylim is given and this is
        not specified then defaults to zero. If xlim is given and
        this is also specified then it is ignored.

    y : scalar or None; default: None
        The y coordinate of the slice. If xlim is given and this is
        not specified then defaults to zero. If ylim is given and
        this is also specified then it is ignored.

    xlim : scalar or size two sequence; default: None
        The limits of the x-axis scan lines. A single number gives
        :math:`x_{\mathrm{axis}} \in [-|x|, +|x|]`, or a tuple of size two gives
        :math:`x_{\mathrm{axis}} \in [x[0], x[1]]`.

    ylim : scalar or array-like; default: None
        The limits of the y-axis scan lines. A single number gives
        :math:`y_{\mathrm{axis}} \in [-|y|, +|y|]`, or a tuple of size two gives
        :math:`y_{\mathrm{axis}} \in [y[0], y[1]]`.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(
        self, name, node, npts, x=None, y=None, xlim=None, ylim=None, w0_scaled=True
    ):
        label = "Pixel intensity"
        unit = r"W m$^{-2}$"

        ScanLine.__init__(
            self, x=x, y=y, xlim=xlim, ylim=ylim, npts=npts, dtype=np.float64
        )
        CCDCamera.__init__(self, name, node, w0_scaled, label=label, unit=unit)

    @property
    def npts(self):
        return super().npts

    @npts.setter
    def npts(self, value):
        super(CCDScanLine, self.__class__).npts.fset(self, value)
        self._update_dtype_shape(self.npts)

    def _get_workspace(self, sim):
        ws = CCDLineWorkspace(self, sim, self._out)
        ws.set_output_fn(ccd_line_output)
        return ws


class CCDPixel(CCDCamera, Pixel):
    """Camera for detecting a single pixel of the beam in terms of the intensity.

    Get the unscaled x and y coordinate data via :attr:`.Pixel.xdata` and
    :attr:`.Pixel.ydata`, respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    x : scalar, optional; default: 0
        The x co-ordinate of the pixel.

    y : scalar, optional; default: 0
        The y co-ordinate of the pixel.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, x=0, y=0, w0_scaled=True):
        label = "Pixel intensity"
        unit = r"W m$^{-2}$"
        Pixel.__init__(self, x, y, dtype=np.float64)
        CCDCamera.__init__(self, name, node, w0_scaled, label=label, unit=unit)

    def _get_workspace(self, sim):
        ws = CameraWorkspace(self, sim)
        ws.set_output_fn(ccd_pixel_output)
        return ws


@float_parameter("f", "Frequency", units="Hz")
class ComplexCamera(Camera, ABC):
    """Abstract type for cameras which detect pixel amplitude and phase.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    f : scalar, optional; default: 0
        Field frequency offset from the carrier to detect.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, f=0, w0_scaled=True):
        if isinstance(self, Image):
            shape = self._x.shape[0], self._y.shape[0]
        elif isinstance(self, ScanLine):
            if self.direction == "x":
                shape = self._x.shape
            else:
                shape = self._y.shape
        elif isinstance(self, Pixel):
            shape = None
        else:
            raise TypeError(
                "Bug detected! ComplexCamera does not derive from Image, "
                "ScanLine or Pixel."
            )

        Camera.__init__(self, name, node, w0_scaled, dtype=np.complex128, shape=shape)

        self.f = f
        self._changing_check = set((self.f,))


@float_parameter("f", "Frequency", units="Hz")
class FieldCamera(ComplexCamera, Image):
    r"""Camera for detecting the full image of the beam in terms of amplitude and phase.

    Get the unscaled x and y coordinate data via :attr:`.Image.xdata` and :attr:`.Image.ydata`,
    respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    xlim : sequence or scalar
        Limits of the x-dimension of the image. If a single number is given then
        this will be computed as :math:`x_{\mathrm{lim}} = [-|x|, +|x|]`.

    ylim : sequence or scalar
        Limits of the y-dimension of the image. If a single number is given then
        this will be computed as :math:`y_{\mathrm{lim}} = [-|y|, +|y|]`.

    npts : int
        Number of points in both axes.

    f : scalar, optional; default: 0
        Field frequency offset from the carrier to detect.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, xlim, ylim, npts, f=0, w0_scaled=True):
        Image.__init__(self, xlim, ylim, npts, dtype=np.complex128)
        ComplexCamera.__init__(self, name, node, f=f, w0_scaled=w0_scaled)

    @property
    def npts(self):
        return super().npts

    @npts.setter
    def npts(self, value):
        super(FieldCamera, self.__class__).npts.fset(self, value)
        self._update_dtype_shape(self.resolution)

    def _get_workspace(self, sim):
        ws = FieldCameraWorkspace(self, sim, self._out)
        ws.set_output_fn(field_camera_output)
        return ws


@float_parameter("f", "Frequency", units="Hz")
class FieldScanLine(ComplexCamera, ScanLine):
    r"""Camera for detecting a slice of the beam in terms of amplitude and phase.

    The :attr:`.ScanLine.direction` (i.e. axis of slice) is determined
    from which of `xlim` or `ylim` is specified.

    Get the unscaled x and y coordinate data via :attr:`.ScanLine.xdata` and
    :attr:`.ScanLine.ydata`, respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    npts : int
        Number of points in slice axis.

    x : scalar or None; default: None
        The x coordinate of the slice. If ylim is given and this is
        not specified then defaults to zero. If xlim is given and
        this is also specified then it is ignored.

    y : scalar or None; default: None
        The y coordinate of the slice. If xlim is given and this is
        not specified then defaults to zero. If ylim is given and
        this is also specified then it is ignored.

    xlim : scalar or size two sequence; default: None
        The limits of the x-axis scan lines. A single number gives
        :math:`x_{\mathrm{axis}} \in [-|x|, +|x|]`, or a tuple of size two gives
        :math:`x_{\mathrm{axis}} \in [x[0], x[1]]`.

    ylim : scalar or array-like; default: None
        The limits of the y-axis scan lines. A single number gives
        :math:`y_{\mathrm{axis}} \in [-|y|, +|y|]`, or a tuple of size two gives
        :math:`y_{\mathrm{axis}} \in [y[0], y[1]]`.

    f : scalar, optional; default: 0
        Field frequency offset from the carrier to detect.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(
        self,
        name,
        node,
        npts,
        x=None,
        y=None,
        xlim=None,
        ylim=None,
        f=0,
        w0_scaled=True,
    ):
        ScanLine.__init__(
            self, x=x, y=y, xlim=xlim, ylim=ylim, npts=npts, dtype=np.complex128
        )
        ComplexCamera.__init__(self, name, node, f=f, w0_scaled=w0_scaled)

    @property
    def npts(self):
        return super().npts

    @npts.setter
    def npts(self, value):
        super(FieldScanLine, self.__class__).npts.fset(self, value)
        self._update_dtype_shape(self.npts)

    def _get_workspace(self, sim):
        ws = FieldLineWorkspace(self, sim, self._out)
        ws.set_output_fn(field_line_output)
        return ws


@float_parameter("f", "Frequency", units="Hz")
class FieldPixel(ComplexCamera, Pixel):
    """Camera for detecting a single pixel of the beam in terms of the amplitude and
    phase.

    Get the unscaled x and y coordinate data via :attr:`.Pixel.xdata` and
    :attr:`.Pixel.ydata`, respectively.

    Parameters
    ----------
    name : str
        Unique name of the camera.

    node : :class:`.OpticalNode`
        Node at which to detect.

    x : scalar, optional; default: 0
        The x co-ordinate of the pixel.

    y : scalar, optional; default: 0
        The y co-ordinate of the pixel.

    f : scalar, optional; default: 0
        Field frequency offset from the carrier to detect.

    w0_scaled : bool, optional; default: True
        Flag indicating whether the :math:`x`, :math:`y` axes
        should be scaled to the waist-size of the beam parameter
        at `node`.
    """

    def __init__(self, name, node, x=0, y=0, f=0, w0_scaled=True):
        Pixel.__init__(self, x, y, dtype=np.complex128)
        ComplexCamera.__init__(self, name, node, f=f, w0_scaled=w0_scaled)

    def _get_workspace(self, sim):
        ws = FieldPixelWorkspace(self, sim)
        ws.set_output_fn(field_pixel_output)
        return ws


def _check_limits(value):
    if isinstance(value, numbers.Number):
        value = [-abs(value), abs(value)]

    elif isinstance(value, np.ndarray):
        value = [value.min(), value.max()]

    elif is_iterable(value):
        if len(value) != 2:
            raise TypeError(
                "Expected limit to be a single number or sequence of size 2 "
                f"but got a sequence of size: {len(value)}"
            )

    else:
        raise TypeError("Unrecognised type for limit value.")

    return value



import pytest

# Register util module assertion error rewriting.
# NOTE: this needs to happen before util is actually imported by any other test code.
pytest.register_assert_rewrite(f"{__name__}.diff")



"""Array parser tests.

Note that expression arrays are tested in `test_expressions.py`.
"""

import re
import pytest
from finesse.script.containers import KatScript, KatElement, KatArray
from finesse.script.exceptions import KatScriptError
from ..util import (
    NAME,
    BOOLEAN,
    NUMBER,
    STRING,
    NONE,
    SPACE,
    COMMA,
    LBRACKET,
    RBRACKET,
)
from ..data import SI_PREFICES, INTEGERS, FLOATS, IMAGINARIES, STRINGS, BOOLEANS


@pytest.mark.parametrize(
    "string,array",
    (
        # Empty.
        ("[]", KatArray(arguments=[], extra=[LBRACKET(1, 18), RBRACKET(1, 19)])),
        (
            "[[]]",
            KatArray(
                arguments=[
                    KatArray(arguments=[], extra=[LBRACKET(1, 19), RBRACKET(1, 20)])
                ],
                extra=[LBRACKET(1, 18), RBRACKET(1, 21)],
            ),
        ),
        (
            "[[[]]]",
            KatArray(
                arguments=[
                    KatArray(
                        arguments=[
                            KatArray(
                                arguments=[], extra=[LBRACKET(1, 20), RBRACKET(1, 21)]
                            )
                        ],
                        extra=[LBRACKET(1, 19), RBRACKET(1, 22)],
                    )
                ],
                extra=[LBRACKET(1, 18), RBRACKET(1, 23)],
            ),
        ),
        (
            "[1, [[]]]",
            KatArray(
                arguments=[
                    NUMBER(1, 19, "1"),
                    KatArray(
                        arguments=[
                            KatArray(
                                arguments=[], extra=[LBRACKET(1, 23), RBRACKET(1, 24)]
                            )
                        ],
                        extra=[LBRACKET(1, 22), RBRACKET(1, 25)],
                    ),
                ],
                extra=[LBRACKET(1, 18), COMMA(1, 20), SPACE(1, 21), RBRACKET(1, 26)],
            ),
        ),
        (
            "[1, [[], 2]]",
            KatArray(
                arguments=[
                    NUMBER(1, 19, "1"),
                    KatArray(
                        arguments=[
                            KatArray(
                                arguments=[], extra=[LBRACKET(1, 23), RBRACKET(1, 24)]
                            ),
                            NUMBER(1, 27, "2"),
                        ],
                        extra=[
                            LBRACKET(1, 22),
                            COMMA(1, 25),
                            SPACE(1, 26),
                            RBRACKET(1, 28),
                        ],
                    ),
                ],
                extra=[LBRACKET(1, 18), COMMA(1, 20), SPACE(1, 21), RBRACKET(1, 29)],
            ),
        ),
        (
            "[0, 0]",
            KatArray(
                arguments=[NUMBER(1, 19, "0"), NUMBER(1, 22, "0")],
                extra=[LBRACKET(1, 18), COMMA(1, 20), SPACE(1, 21), RBRACKET(1, 23)],
            ),
        ),
        (
            "[[1, 0], [0, 1], [0, 2]]",
            KatArray(
                arguments=[
                    KatArray(
                        arguments=[NUMBER(1, 20, "1"), NUMBER(1, 23, "0")],
                        extra=[
                            LBRACKET(1, 19),
                            COMMA(1, 21),
                            SPACE(1, 22),
                            RBRACKET(1, 24),
                        ],
                    ),
                    KatArray(
                        arguments=[NUMBER(1, 28, "0"), NUMBER(1, 31, "1")],
                        extra=[
                            LBRACKET(1, 27),
                            COMMA(1, 29),
                            SPACE(1, 30),
                            RBRACKET(1, 32),
                        ],
                    ),
                    KatArray(
                        arguments=[NUMBER(1, 36, "0"), NUMBER(1, 39, "2")],
                        extra=[
                            LBRACKET(1, 35),
                            COMMA(1, 37),
                            SPACE(1, 38),
                            RBRACKET(1, 40),
                        ],
                    ),
                ],
                extra=[
                    LBRACKET(1, 18),
                    COMMA(1, 25),
                    SPACE(1, 26),
                    COMMA(1, 33),
                    SPACE(1, 34),
                    RBRACKET(1, 41),
                ],
            ),
        ),
        (
            "[[1, 0], [[0, 1]], [[0, 2]]]",
            KatArray(
                arguments=[
                    KatArray(
                        arguments=[NUMBER(1, 20, "1"), NUMBER(1, 23, "0")],
                        extra=[
                            LBRACKET(1, 19),
                            COMMA(1, 21),
                            SPACE(1, 22),
                            RBRACKET(1, 24),
                        ],
                    ),
                    KatArray(
                        arguments=[
                            KatArray(
                                arguments=[NUMBER(1, 29, "0"), NUMBER(1, 32, "1")],
                                extra=[
                                    LBRACKET(1, 28),
                                    COMMA(1, 30),
                                    SPACE(1, 31),
                                    RBRACKET(1, 33),
                                ],
                            )
                        ],
                        extra=[LBRACKET(1, 27), RBRACKET(1, 34)],
                    ),
                    KatArray(
                        arguments=[
                            KatArray(
                                arguments=[NUMBER(1, 39, "0"), NUMBER(1, 42, "2")],
                                extra=[
                                    LBRACKET(1, 38),
                                    COMMA(1, 40),
                                    SPACE(1, 41),
                                    RBRACKET(1, 43),
                                ],
                            )
                        ],
                        extra=[LBRACKET(1, 37), RBRACKET(1, 44)],
                    ),
                ],
                extra=[
                    LBRACKET(1, 18),
                    COMMA(1, 25),
                    SPACE(1, 26),
                    COMMA(1, 35),
                    SPACE(1, 36),
                    RBRACKET(1, 45),
                ],
            ),
        ),
        # Trailing comma.
        pytest.param(
            "[1,]",
            KatArray(
                arguments=[NUMBER(1, 19, "1")],
                extra=[LBRACKET(1, 18), COMMA(1, 20), RBRACKET(1, 21)],
            ),
            id="single-line-trailing-comma-nospace",
        ),
        pytest.param(
            "[1 ,]",
            KatArray(
                arguments=[NUMBER(1, 19, "1")],
                extra=[LBRACKET(1, 18), SPACE(1, 20), COMMA(1, 21), RBRACKET(1, 22)],
            ),
            id="single-line-trailing-comma-prespace",
        ),
        pytest.param(
            "[1, ]",
            KatArray(
                arguments=[NUMBER(1, 19, "1")],
                extra=[LBRACKET(1, 18), COMMA(1, 20), SPACE(1, 21), RBRACKET(1, 22)],
            ),
            id="single-line-trailing-comma-postspace",
        ),
        pytest.param(
            "[1 , ]",
            KatArray(
                arguments=[NUMBER(1, 19, "1")],
                extra=[
                    LBRACKET(1, 18),
                    SPACE(1, 20),
                    COMMA(1, 21),
                    SPACE(1, 22),
                    RBRACKET(1, 23),
                ],
            ),
            id="single-line-trailing-comma-prepostspace",
        ),
    ),
)
def test_array_single_line(parser, string, array):
    assert parser.parse(f"fake_element el1 {string}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[array],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize("number,_a,_b", INTEGERS + FLOATS + IMAGINARIES)
def test_array_numbers(parser, number, _a, _b):
    """Test that numbers as array values parse correctly."""
    assert parser.parse(f"fake_element el1 [{number}]") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatArray(
                        arguments=[NUMBER(1, 19, number)],
                        extra=[LBRACKET(1, 18), RBRACKET(1, 19 + len(number))],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize("prefix,_", SI_PREFICES.items())
def test_array_si_prefix_values(parser, prefix, _):
    """Test that numbers with SI prefices as array values parse correctly."""
    number = f"3.141{prefix}"
    assert parser.parse(f"fake_element el1 [{number}]") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatArray(
                        arguments=[NUMBER(1, 19, number)],
                        extra=[LBRACKET(1, 18), RBRACKET(1, 19 + len(number))],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize("string,_a,_b", STRINGS)
def test_array_string_values(parser, string, _a, _b):
    assert parser.parse(f"fake_element el1 [{string}]") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatArray(
                        arguments=[STRING(1, 19, string)],
                        extra=[LBRACKET(1, 18), RBRACKET(1, 19 + len(string))],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


def test_array_empty_value(parser):
    assert parser.parse("fake_element el1 [none]") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatArray(
                        arguments=[NONE(1, 19)],
                        extra=[LBRACKET(1, 18), RBRACKET(1, 19 + len("none"))],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize("value,_a,_b", BOOLEANS)
def test_array_boolean_values(parser, value, _a, _b):
    assert parser.parse(f"fake_element el1 [{value}]") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[
                    KatArray(
                        arguments=[BOOLEAN(1, 19, value)],
                        extra=[LBRACKET(1, 18), RBRACKET(1, 19 + len(value))],
                    ),
                ],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )


@pytest.mark.parametrize(
    "expression,error",
    (
        pytest.param(
            "[1 1]",
            (
                "\nline 1: array values should be delimited by ','\n"
                "-->1: fake_element el1 [1 1]\n"
                "                         ^"
            ),
            id="no-comma",
        ),
        pytest.param(
            "[1, 1 1]",
            (
                "\nline 1: array values should be delimited by ','\n"
                "-->1: fake_element el1 [1, 1 1]\n"
                "                            ^"
            ),
            id="no-second-comma",
        ),
        pytest.param(
            "[1, (1) 1]",
            (
                "\nline 1: array values should be delimited by ','\n"
                "-->1: fake_element el1 [1, (1) 1]\n"
                "                              ^"
            ),
            id="no-second-comma-expr",
        ),
    ),
)
def test_invalid_syntax(parser, expression, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        parser.parse(f"fake_element el1 {expression}")



"""Convenience objects and functions for unit manipulation."""

import enum

from .misc import find


@enum.unique
class SI(enum.Enum):
    """Enum defining common SI unit constant names. Look-up
    the corresponding values with the `finesse.utilities.SI_MAP`
    dictionary.

    Examples
    --------
    To get the constant for `SI.MILLI` simply look-up:

    >>> SI_MAP[SI.MILLI]
    1e-3
    """

    YOCTO = 0
    ZEPTO = 1
    ATTO = 2
    FEMTO = 3
    PICO = 4
    NANO = 5
    MICRO = 6
    MILLI = 7
    CENTI = 8
    DECI = 9
    NONE = 10
    KILO = 11
    MEGA = 12
    GIGA = 13
    TERA = 14
    PETA = 15


SI_VALUE = {
    SI.YOCTO: 1e-24,
    SI.ZEPTO: 1e-21,
    SI.ATTO: 1e-18,
    SI.FEMTO: 1e-15,
    SI.PICO: 1e-12,
    SI.NANO: 1e-9,
    SI.MICRO: 1e-6,
    SI.MILLI: 1e-3,
    SI.CENTI: 1e-2,
    SI.DECI: 1e-1,
    SI.NONE: 1.0,
    SI.KILO: 1e3,
    SI.MEGA: 1e6,
    SI.GIGA: 1e9,
    SI.TERA: 1e12,
    SI.PETA: 1e15,
}

SI_LABEL = {
    SI.YOCTO: "y",
    SI.ZEPTO: "z",
    SI.ATTO: "a",
    SI.FEMTO: "f",
    SI.PICO: "p",
    SI.NANO: "n",
    SI.MICRO: "u",
    SI.MILLI: "m",
    SI.CENTI: "c",
    SI.DECI: "d",
    SI.KILO: "k",
    SI.MEGA: "M",
    SI.GIGA: "G",
    SI.TERA: "T",
    SI.PETA: "P",
}


def get_SI_value(name):
    """Get an SI value by name.

    Valid names correspond to the ``values()`` of the ``finesse.utilities.SI_LABEL``
    dict.

    Parameters
    ----------
    name : str
        Label name of the SI value (e.g. "m" for milli, "G" for giga).

    Returns
    -------
    value : float
        Value corresponding to the SI name.
    """
    keys = list(SI_LABEL.keys())
    values = list(SI_LABEL.values())

    idx = find(values, name)
    if idx is None:
        raise ValueError(f"Invalid SI name {name}.")

    key = keys[idx]
    return SI_VALUE[key]



"""Test the amplitude of light transmitted through a phase modulator at various frequencies."""

import numpy as np
from scipy.special import jv
import pytest
from finesse import Model
from finesse.components import Laser, Modulator
from finesse.detectors import AmplitudeDetector
from finesse.analysis import xaxis


MAX_MODULATION_ORDER = 5


@pytest.fixture(scope="module")
def modulator_model_out():
    """Fixture with the results of a model with a modulator with changing modulation index.

    See Also
    --------
    3.8.1 in Living Rev Relativ (2016).
    """
    model = Model()
    model.chain(
        Laser("l1", P=1), Modulator("eom1", midx=0, f=40e3, order=5, mod_type="pm")
    )

    for order in range(1, MAX_MODULATION_ORDER + 1):
        f = 40e3 * order
        model.add(AmplitudeDetector(f"bessel{order}", model.eom1.p2.o, f=f))

    return xaxis(model.eom1.midx, "lin", 0, 10, 100)


def test_bessel(modulator_model_out):
    """Test that the transmitted amplitude as a function of modulation index follows a Bessel \
    function."""
    maxdiff = 0
    print('')
    for order in range(1, MAX_MODULATION_ORDER + 1):
        mod_data = np.abs(modulator_model_out[f"bessel{order}"])
        bessel_fcn = np.abs(jv(order, modulator_model_out.x1))
        diff = np.max(np.abs(mod_data - bessel_fcn)/np.abs(bessel_fcn))
        maxdiff = max(maxdiff,diff)
        print(f'Order: {order}, Relative Difference: {maxdiff}')
        assert np.allclose(mod_data, bessel_fcn)

    print('Max relative difference: '+str(maxdiff))

if __name__ == '__main__':
    pytest.main(args=['-v', '-s'])



import pytest
from finesse.script.containers import KatScript, KatElement
from ..util import NAME, SPACE, REFERENCE


@pytest.mark.parametrize("reference", ("&m1.T", "&l1.P", "&mod1.p1.o.q"))
def test_binary_expression(parser, reference):
    assert parser.parse(f"fake_element el1 {reference}") == KatScript(
        arguments=[
            KatElement(
                directive=NAME(1, 1, "fake_element"),
                name=NAME(1, 14, "el1"),
                arguments=[REFERENCE(1, 18, reference)],
                extra=[SPACE(1, 13), SPACE(1, 17)],
            )
        ],
        extra=[],
    )



# -*- coding: utf-8 -*-
"""
Holds the various instances of simulation classes.

Listed below are all the sub-modules of the ``simulations`` module with
a brief description of the contents of each.
"""

# from finesse.simulations.access import AccessSimulation
# from finesse.simulations.base import BaseSimulation
from finesse.simulations.basematrix import CarrierSignalMatrixSimulation
# from finesse.simulations.digraph import DigraphSimulation, DigraphSimulationBase
# from finesse.simulations.debug import DebugSimulation
# from finesse.simulations.dense import DenseSimulation


# __all__ = (
#     "AccessSimulation",
#     "BaseSimulation",
#     "CarrierSignalMatrixSimulation",
#     "DigraphSimulation",
#     "DigraphSimulationBase",
#     "DebugSimulation",
#     "DenseSimulation",
# )



import logging

import numpy as np

from finesse.components.modal.lens import LensWorkspace
from finesse.components.general import Connector, InteractionType
from finesse.components.node import NodeDirection, NodeType
from finesse.parameter import float_parameter
from finesse.utilities import refractive_index
from finesse.components.modal.workspace import KnmConnectorWorkspace

        
@float_parameter(
    "fc",
    "Central frequency",
    units="Hz"
)
@float_parameter(
    "bandwidth",
    "Bandpass bandwidth",
    units="Hz"
)
class OpticalBandpassFilter(Connector):
    """An idealised optical bandpass filter that will transmit an optical frequency around
    some central frequency with a 3dB bandwidth.

    Parameters
    ----------
    name : str
        Name of element
    fc : float, symbol
        Central frequency
    bandwidth : float, symbol
        Filter 3dB bandwidth
    """

    def __init__(self, name, fc=0, bandwidth=1000):
        super().__init__(name)

        self.fc = fc
        self.bandwidth = bandwidth

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # optic to optic couplings
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
        )
        
    def _get_workspace(self, sim):
        from finesse.simulations.basematrix import CarrierSignalMatrixSimulation

        if isinstance(sim, CarrierSignalMatrixSimulation):
            _, is_changing = self._eval_parameters()
            refill = (
                sim.is_component_in_mismatch_couplings(self)
                or len(is_changing)
                or sim.carrier.any_frequencies_changing
                or (sim.signal.any_frequencies_changing if sim.signal else False)
        )
            from .modal.optical_bandpass import optical_bandpass_carrier_fill, optical_bandpass_signal_fill, OpticalBandpassWorkspace

            ws = OpticalBandpassWorkspace(self, sim, refill)
            # This assumes that nr1/nr2 cannot change during a simulation
            ws.nr1 = refractive_index(self.p1)
            ws.nr2 = refractive_index(self.p2)
            # TODO ddb refractive index should be equal on
            # both sides of the lens as we are using the thin
            # lens approximation
            assert ws.nr1 == ws.nr2

            ws.carrier.add_fill_function(optical_bandpass_carrier_fill, refill)
            ws.signal.add_fill_function(optical_bandpass_signal_fill, refill)

            if sim.is_modal:
                # Set the coupling matrix information
                # ABCDs are same in each direction
                ws.set_knm_info(
                    "P1i_P2o",
                    nr_from=ws.nr1,
                    nr_to=ws.nr2,
                    is_transmission=True,
                )
                # TODO nr reversed here for now until it's forced to be same on both sides
                ws.set_knm_info(
                    "P2i_P1o",
                    nr_from=ws.nr2,
                    nr_to=ws.nr1,
                    is_transmission=True,
                )

            return ws
        else:
            raise Exception(f"Optical bandpass filter does not handle a simulation of type {sim}")


"""Legacy parser gauss command tests."""

import pytest


def test_nonexistent_node_invalid(parser):
    """
    Test that creating a gauss command with a component/node that
    doesn't exist throws an error.

    See #221.
    """
    with pytest.raises(Exception):
        parser.parse(
            """
            l laser 2.5 0.0 0.0 n1
            s s1 1 n1 n2
            m m1 0 1 0 n2 n3
            gauss g1 l1 n2 1 10
            noxaxis
            """
        )



import pytest
import finesse
import numpy as np


def test_bs_pitch_yaw_mode_content():
    """
    This reflects a perfectly aligned beam from a bs.
    The beam has a waist at the bs. The bs is misaligned
    by a positive pitch/yaw. As we have right-handed coordinates
    for the components, this means the beam is deflected in a
    negative direction in x/y. Or this means that the 10/01 mode
    content has a negative sign.
    """
    IFO = finesse.Model()

    IFO.parse(
        """
        l l1 P=1
        s s1 l1.p1 m1.p1 L=1e15
        l l2 P=1
        s s2 l2.p1 m1.p2 L=1e15
        bs m1 R=1 T=0 xbeta=1e-9 ybeta=1e-9

        l l3 P=1
        s s3 l3.p1 m1.p3 L=1e15
        l l4 P=1
        s s4 l4.p1 m1.p4 L=1e15

        gauss g1 m1.p1.o z=0 w0=1e-3

        ad u100 l1.p1.i 0 n=0 m=0
        ad u110 l1.p1.i 0 n=1 m=0
        ad u101 l1.p1.i 0 n=0 m=1
        ad u200 l2.p1.i 0 n=0 m=0
        ad u210 l2.p1.i 0 n=1 m=0
        ad u201 l2.p1.i 0 n=0 m=1
        ad u300 l3.p1.i 0 n=0 m=0
        ad u310 l3.p1.i 0 n=1 m=0
        ad u301 l3.p1.i 0 n=0 m=1
        ad u400 l4.p1.i 0 n=0 m=0
        ad u410 l4.p1.i 0 n=1 m=0
        ad u401 l4.p1.i 0 n=0 m=1

        modes(maxtem=1)
        """
    )

    out = IFO.run()

    q = finesse.BeamParam(w0=1e-3, z=0)
    # factor of 2 because angle change is twice the misalignment
    a10 = 2 * IFO.m1.xbeta / q.divergence
    a01 = 2 * IFO.m1.ybeta / q.divergence

    # 00 power should be 1 - the power in ecah 10 01 mode
    assert abs(abs(out["u100"]) ** 2 - (1 - a10 ** 2 - a01 ** 2)) < 1e-15
    assert abs(abs(out["u200"]) ** 2 - (1 - a10 ** 2 - a01 ** 2)) < 1e-15
    assert abs(abs(out["u300"]) ** 2 - (1 - a10 ** 2 - a01 ** 2)) < 1e-15
    assert abs(abs(out["u400"]) ** 2 - (1 - a10 ** 2 - a01 ** 2)) < 1e-15

    # right-handed rotation pitch yaw tilts beam in negative directions
    # after propagation 1e15m it should all be real valued, so a pure displacement
    assert abs(out["u110"] - (-a10)) < 1e-15
    assert abs(out["u101"] - (-a01)) < 1e-15
    assert abs(out["u210"] - (-a10)) < 1e-15
    assert abs(out["u201"] - (-a01)) < 1e-15
    # Yaw other side is same negative sign, but pitch is reversed
    # and beam goes up
    assert abs(out["u310"] - (-a10)) < 1e-15
    assert abs(out["u301"] - (+a01)) < 1e-15
    assert abs(out["u410"] - (-a10)) < 1e-15
    assert abs(out["u401"] - (+a01)) < 1e-15


@pytest.mark.parametrize("port", ("m1.p1", "m1.p2", "m1.p3", "m1.p4"))
def test_z_pitch_yaw_forces_side1(port):
    IFO = finesse.Model()

    IFO.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 {port} L=0
        bs m1 R=1 T=0

        gauss g1 {port}.i z=0 w0=1e-3

        free_mass m1_sus m1.mech mass=inf I_yaw=inf I_pitch=inf

        ad u100 {port}.i 0 n=0 m=0
        ad u110 {port}.i 0 n=1 m=0
        ad u101 {port}.i 0 n=0 m=1

        ad Fz m1.mech.F_z &fsig.f
        ad Fyaw m1.mech.F_yaw &fsig.f
        ad Fpitch m1.mech.F_pitch &fsig.f

        sgen s l1.amp

        fsig({1/(2*np.pi)})
        modes(maxtem=1)
        """
    )
    IFO.l1.tem(1, 0, 1e-3, 0)
    IFO.l1.tem(0, 1, 1e-3, 0)

    q = finesse.BeamParam(w0=1e-3, z=0)

    out = IFO.run()

    dx = out["u110"] * q.w0
    # dy = out["u101"] * q.w0
    side = 1 if ("p1" in port) or ("p2" in port) else 2

    if side == 1:
        # negative here because mirror z direction is opposite to beams
        # so an increase in power applies a neative force
        Fz = -2 * IFO.l1.P / 299792458
        # positive beam x displacements should generate
        # a positive torque on the mirror from right-handed
        # rotation
        Fyaw = dx * 2 * IFO.l1.P / 299792458
        # a positive beam displacement in y should be negative...
        Fpitch = -dx * 2 * IFO.l1.P / 299792458
    else:
        # all opposite from side 2 except yaw
        Fz = 2 * IFO.l1.P / 299792458
        Fyaw = dx * 2 * IFO.l1.P / 299792458
        Fpitch = dx * 2 * IFO.l1.P / 299792458

    assert abs(out["Fz"] - Fz) < 1e-15
    assert abs(out["Fyaw"] - Fyaw) < 1e-15
    assert abs(out["Fpitch"] - Fpitch) < 1e-15


def test_xybeta_curved():
    q = finesse.BeamParam(w0=1e-3, z=100)

    IFO = finesse.Model()
    IFO.parse(
        f"""
        l l1 P=1
        s s1 l1.p1 m1.p1 L={q.z}
        bs m1 R=1 T=0 Rc={q.Rc} xbeta=1e-9 ybeta=1e-9
        s s2 m1.p2 n.p1 L={q.z}
        nothing n

        gauss g1 n.p1.o z=0 w0=1e-3

        ad u100 n.p1.i n=0 m=0 f=0
        ad u110 n.p1.i n=1 m=0 f=0
        ad u101 n.p1.i n=0 m=1 f=0

        modes(maxtem=1)
        """
    )

    q = finesse.BeamParam(w0=1e-3, z=0)

    out = IFO.run()

    dx = out["u110"].real * q.w0
    dy = out["u101"].real * q.w0

    # reflect beam from curved mirror some distance away.
    # back at the waist the beam should be displaced by the lever
    # arm and still tilted the same direction
    assert abs(dx - IFO.m1.xbeta * -2 * IFO.spaces.s1.L) < 1e-14
    assert abs(dy - IFO.m1.ybeta * -2 * IFO.spaces.s1.L) < 1e-14
    # accuracy on numerics here isn't as good
    assert abs(out["u101"].imag * q.divergence - IFO.m1.ybeta) < 1e-9
    assert abs(out["u110"].imag * q.divergence - IFO.m1.xbeta) < 1e-9


@pytest.fixture
def tilted_plate():
    """Transmissive 1m thick plate, refractive index = 2"""
    return """
    l l1 1 0 n0

    s s1 1 1 n0 n1
    bs m1 0 1 0 0 n1 dump n2 dump
    s s2 1 2 n2 n3
    bs m2 0 1 0 0 n3 dump n4 dump
    s s3 0 1 n4 n5

    l l2 1 0 n5
    maxtem 1
    """


def test_tilted_plate_offset_12_yaw(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 1 to 2
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
        {tilted_plate}

        xaxis m1 xbeta lin -1n 1n 1
        put m2 xbeta $x1

        gauss g1 s3 n5 {q.w0} {q.z}
        ad a 1 0 0 n5
        ad c 0 0 0 n5

        yaxis abs:deg
        """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)


def test_tilted_plate_offset_21_yaw(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 2 to 1
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
    {tilted_plate}

    xaxis m1 xbeta lin -1n 1n 1
    put m2 xbeta $x1

    gauss g1 s1 n0 {q.w0} {q.z}
    ad a 1 0 0 n0
    ad c 0 0 0 n0

    yaxis abs:deg
    """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)


def test_tilted_plate_offset_12_pitch(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 1 to 2
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
    {tilted_plate}

    xaxis m1 ybeta lin -1n 1n 1
    put m2 ybeta $x1

    gauss g1 s3 n5 {q.w0} {q.z}
    ad a 0 1 0 n5
    ad c 0 0 0 n5

    yaxis abs:deg
    """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))
    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([-1, 1]) - a) < 1e-13)


def test_tilted_plate_offset_21_pitch(tilted_plate):
    # Tests beam gets displaced by a thick plate at an angle correctly
    # Going side 2 to 1
    IFO = finesse.Model()
    q = finesse.BeamParam(w0=1e-3, z=0)
    IFO.parse_legacy(
        f"""
    {tilted_plate}

    xaxis m1 ybeta lin -1n 1n 1
    put m2 ybeta $x1

    gauss g1 s1 n0 {q.w0} {q.z}
    ad a 0 1 0 n0
    ad c 0 0 0 n0

    yaxis abs:deg
    """
    )

    out = IFO.run()
    a = out["a"] * np.exp(-1j * np.angle(out["c"]))

    delta = (
        IFO.spaces.s2.L * IFO.analysis.stop * (1 - IFO.spaces.s1.nr / IFO.spaces.s2.nr)
    )
    assert all(abs(delta / q.w0 * np.array([1, -1]) - a) < 1e-13)



"""Test cases for model component tree generation."""


def test_component_tree_contains_network_nodes(network_model_michelson):
    """Test that all branches of the tree contain network nodes, regardless of starting node."""
    model, network = network_model_michelson

    for start in (model.L0, model.BS, model.IMX, model.IMY, model.EMX, model.EMY):
        tree = model.component_tree(start)
        assert set([child.name for child in tree.get_all_children()]).issubset(
            set(network.nodes)
        )


def test_component_tree_flattens_cyclic_network(network_model_sagnac):
    """Test that component tree contains a flattened, acyclic version of a cyclic component \
    network.

    Note: cyclic networks are handled by :meth:`networkx.dfs_tree`, so
    :meth:`.TreeNode.from_network` does not need to detect and avoid such cycles.
    """
    model, _ = network_model_sagnac

    for start in (model.L0, model.BS, model.M1, model.M2, model.M3):
        tree = model.component_tree(start)
        assert len(tree.get_all_children()) + 1 == len(model.components)
